/*
 * drivers/char/sid.c - SUZAKU I/O Boards -A/D- Driver
 *
 * Copyright (C) 2006-2008 Atmark Techno, Inc.
 * Chris McHarg <chris (at) atmark-techno.com>
 * Masahiro Hanada <hanada (at) atmark-techno.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * FIXME:
 *  - the driver assume that dma_addr_t is u32.
 */

/* #define DEBUG (1) */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/version.h>
#include <linux/platform_device.h>
#include <linux/cdev.h>
#include <linux/dmapool.h>
#include <linux/sysfs.h>

#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/suzaku_sid.h>

/*
 * Register addresses and setting values
 */
#define REG_CONTROL			(0x0000)
#define REG_CONTROL_SAMPLE		(1 << 0)
#define REG_CONTROL_DMA			(1 << 1)
#define REG_CONTROL_HW_RESET		(1 << 31)

#define REG_SAMPLE_CLOCK		(0x0060)
#define REG_SAMPLE_INTERVAL		(0x0064)
#define REG_DMA_ADDRESS			(0x0080)
#define REG_DMA_LENGTH			(0x0084)

#define REG_STATUS			(0x0010)
#define REG_STATUS_OVERRUN		(1 << 0)
#define REG_STATUS_DMA_DONE		(1 << 1)
#define REG_STATUS_HW_RESET		(1 << 31)

#define REG_STATUS_CLEAR		(0x0014)
#define REG_STATUS_CLEAR_OVERRUN	(1 << 0)
#define REG_STATUS_CLEAR_DMA_DONE	(1 << 1)
#define REG_STATUS_CLEAR_HW_RESET	(1 << 31)

#define REG_INTR_E			(0x0040)
#define REG_INTR_E_OVERRUN		(1 << 0)
#define REG_INTR_E_DMA_DONE		(1 << 1)
#define REG_INTR_E_HW_RESET		(1 << 31)

#define REG_ADC_CHANNELS		(0x00A0)
#define REG_ADC_RESOLUTION		(0x00A4)
#define REG_FIFO_SIZE			(0x00A8)

#define REG_CORE_ID			(0x00FC)

/*
 * Driver specific definitions
 */
#define DRIVER_NAME			"sid"
#define DRIVER_VERSION			"2.0"
#define DRIVER_DESC			"SUZAKU I/O A/D Board Driver"

#ifndef _ALIGN_UP
#define _ALIGN_UP(addr,size)		(((addr)+((size)-1))&(~((size)-1)))
#endif
#ifndef _ALIGN_DOWN
#define _ALIGN_DOWN(addr,size)		((addr)&(~((size)-1)))
#endif

#define MAX_NR_AD_BOARDS		(1)
#define SAMPLE_RATE_MAX			(150000)
#define DMA_TRANS_HZ			(5) //200msec
#define BYTES_PER_CHANNELS(res)		(((res) + 7) / 8)
#define CHANNELS_2_BYTES(ch, res)	(BYTES_PER_CHANNELS(res) * (ch))
#define SRATE_2_BLOCKSIZE(srate, ssize, asize)	\
        /* 16bits/16ch/150000sps/200msec => 960000Bytes */ \
        ({u32 n = (ssize) * (((srate) + (DMA_TRANS_HZ - 1)) / DMA_TRANS_HZ) \
                  + (asize); \
          (n > PAGE_SIZE) ? _ALIGN_DOWN(PAGE_ALIGN(n) - (asize), (ssize)) \
                          : (n - (asize));})
#define BYTES_2_SAMPLES(n, ssize)	((n) / (ssize))
#define DMA_POOL_ALLOC_MAX		(10)
#define DMA_MEMORY_SIZE			(0x00200000)
#define SRATE_2_SCYCLE(c, r)		(((c) + ((r) >> 1)) / (r))

#define dev_func(dev) \
	dev_dbg(dev, "%s()\n", __FUNCTION__)

#define resource_start(res) \
	((res)->start)

#define resource_len(res) \
	((res)->end - (res)->start + 1)

#define COREID(id) \
	(((id) >> 16) & 0xffff)
#define COREVER(id) \
	(((id) >> 4) & 0x0fff)

struct dma_block {
	struct list_head list; /* size: 8bytes */

	u8 *dma_address;
	u32 length;
	u32 overrun_period;

	dma_addr_t handle;
	dma_addr_t dma_address_phys;

	u32 pad[1];

	/* Data must be align to 16 bytes */;
	u8 data[0];
};

struct sid {
	struct device *dev;
	struct cdev cdev;
	dev_t devno;

	int use_count;

	int irq;
	void __iomem *addr;
	unsigned long region_start;
	unsigned long region_len;

	u32 sample_clock;
	u32 adc_channels;
	u32 adc_resolution;
	u32 fifo_size;

	u32 sample_size;
	u32 sample_rate;
	u32 sample_cycle;

	u32 overrun_period_max;
	u32 overrun_jiffies_base;
	u32 overrun_adjust;

	struct list_head dma_blocks;
	struct list_head read_blocks;

	struct dma_pool *dma_pool;
	size_t dma_pool_size;

	spinlock_t use_count_lock;
	spinlock_t dma_blocks_lock;
	wait_queue_head_t read_wait;

	struct dma_block *current_dma_block;
	struct dma_block *next_dma_block;

	unsigned long overrun_jiffies;
	int overrun_flag;

	u32 overrun;
};

static struct sid *sids[MAX_NR_AD_BOARDS];
static u32 default_sample_rate = 10000;
module_param(default_sample_rate, uint, 10000);
MODULE_PARM_DESC(default_sample_rate, "Default sample rate");

#ifdef DEBUG
static inline u32
_read_reg(struct sid *sid, u32 offset)
{
	if (!sid->addr) {
		BUG();
		return 0;
	}
	return __raw_readl(sid->addr + offset);
}

static inline void
_write_reg(struct sid *sid, u32 offset, u32 val)
{
	if (!sid->addr) {
		BUG();
		return;
	}
	__raw_writel(val, sid->addr + offset);
}
#else
#define _read_reg(sid, offset) \
        __raw_readl((sid)->addr + (offset))
#define _write_reg(sid, offset, val) \
        __raw_writel((val), (sid)->addr + (offset))
#endif

struct id_struct {
	u16 id;
	u16 version;
	u16 version_mask;
};

#define CORE_ID_AND_VERSION(_id, _ver, _mask) \
{ \
	.id = (_id), \
	.version = (_ver), \
	.version_mask = (_mask), \
}

static const struct id_struct sid_ids[] = {
	CORE_ID_AND_VERSION(0x5009, 0x0300, 0xfff0),
	CORE_ID_AND_VERSION(0x5009, 0x0310, 0xfff0),
};

static inline int
is_support_core(u32 id)
{
	u16 core_id = COREID(id);
	u16 core_ver = COREVER(id);
	int i;

	for (i = 0; i < ARRAY_SIZE(sid_ids); i++)
		if ((sid_ids[i].id == core_id) &&
		    (sid_ids[i].version ==
		     (core_ver & sid_ids[i].version_mask)))
			return 0;

	return -ENODEV;
}

static inline void
set_sid(struct sid *sid, int minor)
{
	sids[minor] = sid;
}

static inline struct sid *
get_sid(int minor)
{
	return sids[minor];
}

static int
sid_destroy_dma_pool(struct sid *sid)
{
	struct dma_block *blk_cursor, *blk_next;

	dev_func(sid->dev);

	if (sid->dma_pool) {
		sid->current_dma_block = NULL;
		sid->next_dma_block = NULL;

		list_for_each_entry_safe(blk_cursor, blk_next,
					 &sid->read_blocks, list) {
			list_del(&blk_cursor->list);
			dma_pool_free(sid->dma_pool, blk_cursor,
				      blk_cursor->handle);
		}
		list_for_each_entry_safe(blk_cursor, blk_next,
					 &sid->dma_blocks, list) {
			list_del(&blk_cursor->list);
			dma_pool_free(sid->dma_pool, blk_cursor,
				      blk_cursor->handle);
		}

		/* DMA pool destroy */
		dma_pool_destroy(sid->dma_pool);
		dev_dbg(sid->dev, "dma_pool_destroy\n");
		sid->dma_pool = NULL;
	}

	return 0;
}

static int
sid_create_dma_pool(struct sid *sid)
{
	int dma_pool_alloc_num;
	dma_addr_t handle;
	struct dma_block *dma_block;
	int i;

	dev_func(sid->dev);

	sid_destroy_dma_pool(sid);

	/* DMA pool create */
	sid->dma_pool_size = SRATE_2_BLOCKSIZE(sid->sample_rate,
					       sid->sample_size,
					       sizeof(struct dma_block));
	dev_dbg(sid->dev, "dma_pool_create: size = %d (%d samples)\n",
		sid->dma_pool_size,
		BYTES_2_SAMPLES(sid->dma_pool_size, sid->sample_size));
	sid->dma_pool = dma_pool_create(DRIVER_NAME, sid->dev,
					sizeof(struct dma_block) +
					sid->dma_pool_size,
					4, 0);
	if (!sid->dma_pool) {
		dev_err(sid->dev, "Failed to create dma pool.\n");
		return -ENOMEM;
	}

	dma_pool_alloc_num = DMA_MEMORY_SIZE /
	  PAGE_ALIGN(sizeof(struct dma_block) + sid->dma_pool_size);
	if (dma_pool_alloc_num > DMA_POOL_ALLOC_MAX)
		dma_pool_alloc_num = DMA_POOL_ALLOC_MAX;
	for (i = 0; i < dma_pool_alloc_num; i++) {
		dma_block = dma_pool_alloc(sid->dma_pool, GFP_ATOMIC, &handle);
		if (!dma_block)
			break;

		dma_block->handle = handle;
		dma_block->dma_address_phys =
		  (dma_addr_t)((struct dma_block *)dma_block->handle)->data;
		list_add_tail(&dma_block->list, &sid->dma_blocks);

		dev_dbg(sid->dev, "dma_pool_alloc: dma_size: %d\n",
			sid->dma_pool_size);
	}
	if (list_empty(&sid->dma_blocks)) /* Min 1 block */
		return -ENOMEM;
	if (i < dma_pool_alloc_num)
		dev_warn(sid->dev, "Attempted to allocate %d blocks, "
			 "but only got %d.\n", dma_pool_alloc_num, i);

	_write_reg(sid, REG_DMA_LENGTH,
		   BYTES_2_SAMPLES(sid->dma_pool_size, sid->sample_size));

	return 0;
}

static inline void
sid_start_current_dma(struct sid *sid)
{
	dev_func(sid->dev);

	/* DMA start */
	_write_reg(sid, REG_CONTROL, REG_CONTROL_DMA | REG_CONTROL_SAMPLE);

	if (!sid->overrun_flag)
		sid->current_dma_block->overrun_period = 0;
	else {
		/* DMA start jiffies - Overrun occured jiffies */
		sid->current_dma_block->overrun_period =
		  jiffies - sid->overrun_jiffies;
		sid->overrun_flag = 0;
	}
}

static inline void
sid_start_dma(struct sid *sid, struct dma_block *block)
{
	dev_func(sid->dev);

	sid->current_dma_block = block;

	_write_reg(sid, REG_DMA_ADDRESS, block->dma_address_phys);

	sid_start_current_dma(sid);

	block->dma_address = block->data;
	block->length = sid->dma_pool_size;

	/* Prepare next dma block */
	if (block->list.next != &sid->dma_blocks)
		sid->next_dma_block =
		  list_entry(block->list.next, typeof(struct dma_block), list);
	else
		sid->next_dma_block = NULL;
}

static inline void
sid_complete_dma(struct sid *sid, struct dma_block *block)
{
	dev_func(sid->dev);

	/* Move DMA -> read */
	list_move_tail(&block->list, &sid->read_blocks);
}

static inline void
sid_restore_dma(struct sid *sid, struct dma_block *block)
{
	dev_func(sid->dev);

	/* Move read -> DMA */
	list_move_tail(&block->list, &sid->dma_blocks);
}

static inline int
sid_start(struct sid *sid)
{
	int ret;

	dev_func(sid->dev);

	sid->overrun_period_max = (0xffffffff - (HZ >> 1)) / sid->sample_rate;
	sid->overrun_jiffies_base =
	  (sid->fifo_size * HZ - ((sid->sample_size * HZ) >> 1) +
	   (sid->sample_size * sid->sample_rate - 1)) /
	  (sid->sample_size * sid->sample_rate);
	sid->overrun_adjust =
	  (sid->overrun_jiffies_base * sid->sample_size * sid->sample_rate -
	   sid->fifo_size * HZ + ((sid->sample_size * HZ) >> 1)) /
	  (sid->sample_size * HZ);

	sid->overrun_flag = 0;
	sid->overrun = 0;

	ret = sid_create_dma_pool(sid);
	if (ret < 0)
		return ret;

	_write_reg(sid, REG_SAMPLE_INTERVAL, sid->sample_cycle - 1);

	sid_start_dma(sid, list_entry(sid->dma_blocks.next,
				      typeof(struct dma_block), list));

	return 0;
}

static inline void
sid_stop(struct sid *sid)
{
	dev_func(sid->dev);

	/* DMA stop and FIFO clear */
	_write_reg(sid, REG_CONTROL, 0);
	_write_reg(sid, REG_STATUS_CLEAR,
		   REG_STATUS_CLEAR_DMA_DONE | REG_STATUS_CLEAR_OVERRUN);
	_write_reg(sid, REG_STATUS_CLEAR, 0);
}

static int
sid_properly_sample_rate(struct sid *sid, u32 *rate, u32 *cycle)
{
	dev_func(sid->dev);

	if (!*rate || *rate > SAMPLE_RATE_MAX) {
		dev_err(sid->dev,
			"Given sample rate %uHz is not supported.\n", *rate);
		return -EINVAL;
	}

	*cycle = SRATE_2_SCYCLE(sid->sample_clock, *rate);
	if (sid->sample_clock % *rate) {
		dev_warn(sid->dev,
			 "Given sample rate %uHz is not acceptable. "
			 "Using %u.%04u%04uHz.\n",
			 *rate, sid->sample_clock / *cycle,
			 (((sid->sample_clock % *cycle) * 10000) / *cycle),
			 ((((sid->sample_clock % *cycle) * 10000) % *cycle) *
			  10000) / *cycle);

		/* Update rate */
		*rate = (sid->sample_clock + (*cycle >> 1)) / *cycle;
	}

	return 0;
}

static inline void
sid_disable_interrupt(struct sid *sid)
{
	_write_reg(sid, REG_INTR_E, 0);
}

static inline void
sid_enable_interrupt(struct sid *sid)
{
	_write_reg(sid, REG_STATUS_CLEAR,
		   REG_STATUS_CLEAR_DMA_DONE | REG_STATUS_CLEAR_OVERRUN);
	_write_reg(sid, REG_STATUS_CLEAR, 0);
	_write_reg(sid, REG_INTR_E, REG_INTR_E_DMA_DONE | REG_INTR_E_OVERRUN);
}

static int
sid_set_sample_rate(struct sid *sid, u32 rate)
{
	u32 cycle;
	int ret;

	dev_func(sid->dev);

	ret = sid_properly_sample_rate(sid, &rate, &cycle);
	if (ret < 0)
		return ret;

	sid_disable_interrupt(sid);

	sid_stop(sid);

	sid->sample_rate = rate;
	sid->sample_cycle = cycle;
	dev_dbg(sid->dev, "sample_rate:    0x%08x\n", sid->sample_rate);
	dev_dbg(sid->dev, "sample_cycle:   0x%08x\n", sid->sample_cycle);

	sid_enable_interrupt(sid);

	ret = sid_start(sid);
	if (ret < 0)
		return ret;

	return 0;
}

static int
sid_hw_reset(struct sid *sid)
{
	int timeout = 1000;

	dev_func(sid->dev);

	_write_reg(sid, REG_CONTROL, REG_CONTROL_HW_RESET);
	while (timeout--) {
		if (_read_reg(sid, REG_STATUS) & REG_STATUS_HW_RESET)
			break;

		msleep(1);
	}

	if (timeout <= 0) {
		dev_err(sid->dev, "Timeout waiting for reset.\n");
		return -EFAULT;
	}

	_write_reg(sid, REG_STATUS_CLEAR, REG_STATUS_CLEAR_HW_RESET);
	_write_reg(sid, REG_STATUS_CLEAR, 0);

	return 0;
}

static int
sid_reset(struct sid *sid)
{
	int ret;

	dev_func(sid->dev);

	ret = sid_hw_reset(sid);
	if (ret < 0)
		return ret;

	sid_enable_interrupt(sid);

	ret = sid_start(sid);
	if (ret < 0)
		return ret;

	return 0;
}

static int
sid_setup(struct sid *sid)
{
	int ret;

	dev_func(sid->dev);

	spin_lock_init(&sid->dma_blocks_lock);
	INIT_LIST_HEAD(&sid->dma_blocks);
	INIT_LIST_HEAD(&sid->read_blocks);

	sid->sample_clock = _read_reg(sid, REG_SAMPLE_CLOCK);
	sid->adc_channels = _read_reg(sid, REG_ADC_CHANNELS);
	sid->adc_resolution = _read_reg(sid, REG_ADC_RESOLUTION);
	sid->fifo_size = _read_reg(sid, REG_FIFO_SIZE);
	if (!sid->sample_clock || !sid->adc_channels || !sid->adc_resolution)
		return -ENODEV;

	sid->sample_size =
	  CHANNELS_2_BYTES(sid->adc_channels, sid->adc_resolution);
	sid->sample_rate = default_sample_rate;
	sid->sample_cycle =
	  SRATE_2_SCYCLE(sid->sample_clock, default_sample_rate);

	dev_dbg(sid->dev, "sample_clock:   0x%08x\n", sid->sample_clock);
	dev_dbg(sid->dev, "adc_channels:   0x%08x\n", sid->adc_channels);
	dev_dbg(sid->dev, "adc_resolution: 0x%08x\n", sid->adc_resolution);
	dev_dbg(sid->dev, "sample_size:    0x%08x\n", sid->sample_size);
	dev_dbg(sid->dev, "sample_rate:    0x%08x\n", sid->sample_rate);
	dev_dbg(sid->dev, "sample_cycle:   0x%08x\n", sid->sample_cycle);

	ret = sid_reset(sid);
	if (ret < 0)
		return ret;

	return 0;
}

static irqreturn_t
sid_intr(int irq, void *dev_id, struct pt_regs *regs)
{
	struct sid *sid = dev_id;
	u32 status;

	dev_func(sid->dev);

	status = _read_reg(sid, REG_STATUS);

	if (status & REG_STATUS_DMA_DONE) {
		unsigned long flags;

		_write_reg(sid, REG_STATUS_CLEAR, REG_STATUS_CLEAR_DMA_DONE);
		_write_reg(sid, REG_STATUS_CLEAR, 0);

		spin_lock_irqsave(&sid->dma_blocks_lock, flags);
		sid_complete_dma(sid, sid->current_dma_block);
		if (sid->next_dma_block)
			sid_start_dma(sid, sid->next_dma_block);
		else
			sid->current_dma_block = NULL;
		spin_unlock_irqrestore(&sid->dma_blocks_lock, flags);

		if (waitqueue_active(&sid->read_wait))
			wake_up(&sid->read_wait);
	}

	if (unlikely(status & REG_STATUS_OVERRUN)) {
		/* Overrun occured jiffies */
		sid->overrun_jiffies = jiffies - sid->overrun_jiffies_base;
		sid->overrun_flag = 1;

		sid_stop(sid);

		if (sid->current_dma_block)
			sid_start_current_dma(sid);

#ifdef  DEBUG
		dev_warn(sid->dev, "Hardware fifo overrun detected.\n");
#endif
	}

	dev_dbg(sid->dev, "status: %08x %s%s\n", status,
		((status & REG_STATUS_OVERRUN) ? "[OVERRUN]" : ""),
		((status & REG_STATUS_DMA_DONE) ? "[DMA_DONE]" : ""));

	return IRQ_HANDLED;
}

static ssize_t
sid_read(struct file *filp, char __user *buf, size_t count, loff_t *f_pos)
{
	struct sid *sid = filp->private_data;
	size_t remain, len;
	struct dma_block *blk_cursor, *blk_next;
	int ret;

	dev_func(sid->dev);

	len = count % sid->sample_size;
	if (len)
		count -= len;
	if (!count)
		return 0;
	remain = count;

	sid->overrun = 0;
	for (; ; ) {
		list_for_each_entry_safe(blk_cursor, blk_next,
					 &sid->read_blocks, list) {
			if (blk_cursor->overrun_period) {
				if (remain != count)
					/* Discrete data by overrun */
					return count - remain;

				if (blk_cursor->overrun_period <=
				    sid->overrun_period_max)
					/* Overrun period -> Lost samples */
					sid->overrun =
					  (blk_cursor->overrun_period *
					   sid->sample_rate + (HZ >> 1)) / HZ -
					  sid->overrun_adjust;
				else
					/* Lost many samples */
					sid->overrun = 0xffffffffUL;
				blk_cursor->overrun_period = 0;
			}

			if (blk_cursor->length <= remain)
				len = blk_cursor->length;
			else
				len = remain;

			ret = copy_to_user(buf, blk_cursor->dma_address, len);
			buf += len;

			blk_cursor->dma_address += len;
			blk_cursor->length -= len;
			if (!blk_cursor->length) {
				unsigned long flags;

				spin_lock_irqsave(&sid->dma_blocks_lock,
						  flags);
				sid_restore_dma(sid, blk_cursor);
				if (!sid->current_dma_block)
					sid_start_dma(sid, blk_cursor);
				else if (!sid->next_dma_block)
					sid->next_dma_block = blk_cursor;
				spin_unlock_irqrestore(&sid->dma_blocks_lock,
						       flags);
			}

			if (ret) {
				/* 'copy_to_user' failed */
				dev_err(sid->dev, "Failed to copy data.\n");
				return -EFAULT;
			}

			remain -= len;
			if (!remain)
				/* Read complete */
				return count;
		}
		if (filp->f_flags & O_NONBLOCK)
			/* Non block break */
			return count - remain;

		dev_dbg(sid->dev, "Sleep on...\n");
		ret = wait_event_interruptible(sid->read_wait,
					       !list_empty(&sid->read_blocks));
		dev_dbg(sid->dev, "Wake up: ret = %d\n", ret);
		if (ret < 0)
			/* Cancel by signal */
			return ret;
	}
}

static int
sid_ioctl(struct inode *inode, struct file *filp,
	  unsigned int cmd, unsigned long arg)
{
	struct sid *sid = filp->private_data;

	dev_func(sid->dev);

	switch(cmd) {
	case SID_IOC_GET_CHANNELS:
		return put_user(sid->adc_channels, (int *)arg);

	case SID_IOC_GET_RESOLUTION:
		return put_user(sid->adc_resolution, (int *)arg);

	case SID_IOC_GET_OVERRUN:
		return put_user(sid->overrun, (int *)arg);

	case SID_IOC_SET_FREQ:
		return sid_set_sample_rate(sid, (unsigned int)arg);

	case SID_IOC_RESET:
		return sid_reset(sid);

	default:
		return -EINVAL;
	}
}

static int
sid_open(struct inode *inode, struct file *filp)
{
	int minor = iminor(inode);
	struct sid *sid = get_sid(minor);
	int ret;
	unsigned long flags;

	dev_func(sid->dev);

	if (!sid)
		return -ENODEV;

	spin_lock_irqsave(&sid->use_count_lock, flags);
	if (sid->use_count) {
		spin_unlock_irqrestore(&sid->use_count_lock, flags);
		return -EBUSY;
	}
	sid->use_count++;
	spin_unlock_irqrestore(&sid->use_count_lock, flags);

	filp->private_data = sid;

	ret = request_irq(sid->irq, sid_intr, SA_SHIRQ, DRIVER_NAME, sid);
	if (ret < 0) {
		dev_err(sid->dev, "Failed to obtain IRQ.\n");
		goto err_request_irq;
	}

	ret = sid_setup(sid);
	if (ret < 0)
		goto err_setup;

	return 0;

 err_setup:
	free_irq(sid->irq, sid);
 err_request_irq:
	spin_lock_irqsave(&sid->use_count_lock, flags);
	sid->use_count--;
	spin_unlock_irqrestore(&sid->use_count_lock, flags);

	return ret;
}

static int
sid_release(struct inode *inode, struct file *filp)
{
	int minor = iminor(inode);
	struct sid *sid = get_sid(minor);
	unsigned long flags;

	dev_func(sid->dev);

	sid_hw_reset(sid);

	sid->overrun_flag = 0;
	sid->overrun = 0;

	sid_destroy_dma_pool(sid);

	free_irq(sid->irq, sid);

	spin_lock_irqsave(&sid->use_count_lock, flags);
	sid->use_count--;
	spin_unlock_irqrestore(&sid->use_count_lock, flags);

	return 0;
}

static struct file_operations sid_fops = {
	.owner	 = THIS_MODULE,
	.read	 = sid_read,
	.ioctl	 = sid_ioctl,
	.open	 = sid_open,
	.release = sid_release,
};

static ssize_t
sid_show_core_id(struct device *dev,
		 struct device_attribute *attr, char *buf)
{
	struct sid *sid = dev->driver_data;
	u32 core_id = _read_reg(sid, REG_CORE_ID);
	return snprintf(buf, PAGE_SIZE, "%#hx:%#hx\n",
			COREID(core_id), COREVER(core_id));
}

static ssize_t
sid_show_sample_clock(struct device *dev,
		      struct device_attribute *attr, char *buf)
{
	struct sid *sid = dev->driver_data;
	return snprintf(buf, PAGE_SIZE, "%d\n",
			_read_reg(sid, REG_SAMPLE_CLOCK));
}

static ssize_t
sid_show_channels(struct device *dev,
		  struct device_attribute *attr, char *buf)
{
	struct sid *sid = dev->driver_data;
	return snprintf(buf, PAGE_SIZE, "%d\n",
			_read_reg(sid, REG_ADC_CHANNELS));
}

static ssize_t
sid_show_resolution(struct device *dev,
		    struct device_attribute *attr, char *buf)
{
	struct sid *sid = dev->driver_data;
	return snprintf(buf, PAGE_SIZE, "%d\n",
			_read_reg(sid, REG_ADC_RESOLUTION));
}

static ssize_t
sid_show_overrun(struct device *dev,
		 struct device_attribute *attr, char *buf)
{
	struct sid *sid = dev->driver_data;
	return snprintf(buf, PAGE_SIZE, "%d\n", sid->overrun);
}

static ssize_t
sid_show_default_sample_rate(struct device *dev,
			     struct device_attribute *attr, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "%u\n", default_sample_rate);
}

static ssize_t
sid_store_default_sample_rate(struct device *dev,
			      struct device_attribute *attr,
			      const char *buf, size_t count)
{
	struct sid *sid = dev->driver_data;
	u32 rate, cycle;
	int ret;

	sscanf(buf, "%u", &rate);

	if (rate == default_sample_rate)
		return strnlen(buf, PAGE_SIZE);

	ret = sid_properly_sample_rate(sid, &rate, &cycle);
	if (ret < 0)
		return ret;

	default_sample_rate = rate;
	return strnlen(buf, PAGE_SIZE);
}

static DEVICE_ATTR(core_id, S_IRUGO, sid_show_core_id, NULL);
static DEVICE_ATTR(sample_clock, S_IRUGO, sid_show_sample_clock, NULL);
static DEVICE_ATTR(channels, S_IRUGO, sid_show_channels, NULL);
static DEVICE_ATTR(resolution, S_IRUGO, sid_show_resolution, NULL);
static DEVICE_ATTR(overrun, S_IRUGO, sid_show_overrun, NULL);
static DEVICE_ATTR(default_sample_rate, S_IRUGO | S_IWUSR,
		   sid_show_default_sample_rate,
		   sid_store_default_sample_rate);

static int
sid_probe(struct platform_device *pdev)
{
	struct sid *sid;
	struct resource *res_mem, *res_irq, *res;
	u32 core_id;
	int sid_minor = 0;
	int ret;

	dev_func(&pdev->dev);

	sid = kzalloc(sizeof(struct sid), GFP_KERNEL);
	if (!sid) {
		dev_err(&pdev->dev, "Failed to obtain memory for device.\n");
		ret = -ENOMEM;
		goto err_sid_alloc;
	}
	init_waitqueue_head(&sid->read_wait);

	sid->dev = &pdev->dev;

	res_mem = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res_mem) {
		dev_err(sid->dev, "Failed to get memory region resource.\n");
		ret = -ENODEV;
		goto err_get_res_mem;
	}

	res_irq = platform_get_resource(pdev, IORESOURCE_IRQ, 0);
	if (!res_irq) {
		dev_err(sid->dev, "Failed to get IRQ resource.\n");
		ret = -ENODEV;
		goto err_get_res_irq;
	}

	sid->irq		= resource_start(res_irq);
	sid->region_start	= resource_start(res_mem);
	sid->region_len		= resource_len(res_mem);
	res = request_mem_region(sid->region_start, sid->region_len,
				 DRIVER_NAME);
	if (!res) {
		dev_err(sid->dev,
			"Failed to obtain device I/O memory region.\n");
		ret = -ENOMEM;
		goto err_request_mem_region;
	}
	sid->addr = (void __iomem *)ioremap_nocache(sid->region_start,
						    sid->region_len);
	if (!sid->addr) {
		dev_err(sid->dev,
			"Failed to remap device I/O memory region.\n");
		ret = -EBUSY;
		goto err_ioremap;
	}

	core_id = _read_reg(sid, REG_CORE_ID);
	ret = is_support_core(core_id);
	if (ret < 0) {
		dev_err(sid->dev, "Unsupported SID. (id: %#hx; ver: %#hx)\n",
			COREID(core_id), COREVER(core_id));
		goto err_unsupport;
	}

	sid->devno = MKDEV(SID_MAJOR, sid_minor);
	ret = register_chrdev_region(sid->devno, 1, DRIVER_NAME);
	if (ret < 0) {
		dev_err(sid->dev, "Failed to obtain major number.\n");
		goto err_register_chrdev;
	}
	cdev_init(&sid->cdev, &sid_fops);
	sid->cdev.owner = THIS_MODULE;
	ret = cdev_add(&sid->cdev, sid->devno, 1);
	if (ret < 0) {
		dev_err(sid->dev,
			"Failed to register as a character device.\n");
		goto err_cdev_add;
	}

	sid->dev->driver_data = sid;
	device_create_file(sid->dev, &dev_attr_core_id);
	device_create_file(sid->dev, &dev_attr_sample_clock);
	device_create_file(sid->dev, &dev_attr_channels);
	device_create_file(sid->dev, &dev_attr_resolution);
	device_create_file(sid->dev, &dev_attr_overrun);
	device_create_file(sid->dev, &dev_attr_default_sample_rate);

	sid->use_count = 0;
	set_sid(sid, sid_minor);

	spin_lock_init(&sid->use_count_lock);
	dev_info(sid->dev, "Detected SID-Core. (%#hx; %#hx)\n",
		 COREID(core_id), COREVER(core_id));

	return 0;

 err_cdev_add:
	unregister_chrdev_region(sid->devno, 1);
 err_register_chrdev:
 err_unsupport:
	iounmap(sid->addr);
 err_ioremap:
	release_mem_region(sid->region_start, sid->region_len);
 err_request_mem_region:
 err_get_res_irq:
 err_get_res_mem:
	kfree(sid);
 err_sid_alloc:
	return ret;
}

static int
sid_remove(struct platform_device *pdev)
{
	struct sid *sid;
	int sid_minor = 0;

	dev_func(&pdev->dev);

	sid = get_sid(sid_minor);
	set_sid(NULL, sid_minor);
	if (sid) {
		device_remove_file(sid->dev, &dev_attr_core_id);
		device_remove_file(sid->dev, &dev_attr_sample_clock);
		device_remove_file(sid->dev, &dev_attr_channels);
		device_remove_file(sid->dev, &dev_attr_resolution);
		device_remove_file(sid->dev, &dev_attr_overrun);
		device_remove_file(sid->dev, &dev_attr_default_sample_rate);
		cdev_del(&sid->cdev);
		unregister_chrdev_region(sid->devno, 1);
		iounmap(sid->addr);
		release_mem_region(sid->region_start, sid->region_len);
		kfree(sid);
	}

	return 0;
}

static struct platform_driver sid_driver = {
	.driver	= {
		.name	= DRIVER_NAME,
		.bus	= &platform_bus_type,
	},
	.probe	= sid_probe,
	.remove = sid_remove,
};

static int __init
sid_init(void)
{
	memset(sids, 0, sizeof(struct sid *) * MAX_NR_AD_BOARDS);

	return platform_driver_register(&sid_driver);
}

static void __exit
sid_cleanup(void)
{
	platform_driver_unregister(&sid_driver);
}

module_init(sid_init);
module_exit(sid_cleanup);

MODULE_AUTHOR("Atmark Techno, Inc.");
MODULE_LICENSE("GPL v2");
MODULE_VERSION(DRIVER_VERSION);
MODULE_DESCRIPTION("SUZAKU I/O Boards -A/D- Driver");
