/**
 * drivers/char/sil-7seg-cgi.c
 *     -- SUZAKU I/O Board -LED/SW- 7-Segment-LED Driver for CGI demo
 *
 *  Copyright (C) 2006-2007 Atmark Techno, Inc.
 *
 * Tetsuya OHKAWA <tetsuya@atmark-techno.com>
 * Masahiro Hanada <hanada@atmark-techno.com>
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>
#include <asm/suzaku_sil.h>

#define DEV_NAME                 "sil7segc"
#define DEV_MAJOR                65
#define DEV_VER                  "1.0.1"

#define SIL_7SEG_MAX             (3)
#define SIL_7SEG_CGI_BASEADDR    XPAR_SIL_CNTLR_BASEADDR
#define SIL_7SEG_CGI_REMAP_SIZE  (XPAR_SIL_CNTLR_HIGHADDR - XPAR_SIL_CNTLR_BASEADDR + 1)

static u8 *led_addr;

static struct semaphore sem;

static u8 read_led(int nr)
{
	return __raw_readb(led_addr + nr);
}

static void write_led(int nr, u8 value)
{
	__raw_writeb(value, led_addr + nr);
}

static ssize_t led_7segc_read(struct file *filp, char __user *buf,
			      size_t count, loff_t *offp)
{
	ssize_t retval = 0;
	u8 valbuf[SIL_7SEG_MAX];
	int  i;

	if (down_interruptible(&sem)) {
		retval = -ERESTARTSYS;
		goto err;
	}

	if (count < SIL_7SEG_MAX) {
		retval = -EINVAL;
		goto err;
	}

	if (buf == NULL) {
		retval = -EFAULT;
		goto err;
	}

	for (i = 0; i < SIL_7SEG_MAX; i++)
		valbuf[i] = read_led(i) & 0xf;

	if (copy_to_user(buf, valbuf, SIL_7SEG_MAX)) {
		retval = -EFAULT;
		goto err;
	}
	retval = SIL_7SEG_MAX;

err:
	up(&sem);
	return retval;
}

static ssize_t led_7segc_write(struct file *filp, const char __user *buf,
			       size_t count, loff_t *offp)
{
	ssize_t retval = 0;
	u8      valbuf[SIL_7SEG_MAX];
	int     i;

	if (down_interruptible(&sem)) {
		retval = -ERESTARTSYS;
		goto err;
	}

	if (count < SIL_7SEG_MAX) {
		goto err;
	}
	count = SIL_7SEG_MAX;

	if (buf == NULL) {
		retval = -EFAULT;
		goto err;
	}

	if (copy_from_user(valbuf, buf, count)) {
		retval = -EFAULT;
		goto err;
	}

	for (i = 0; i < count; i++)
		write_led(i, valbuf[i]);

	retval = count;
err:
	up(&sem);
	return retval;
}

static int led_7segc_open(struct inode *inode, struct file *filp)
{
	return 0;
}

static int led_7segc_release(struct inode *inode, struct file *filp)
{
	return 0;
}

struct file_operations led_7segc_fops = {
	owner:    THIS_MODULE,
	read:     led_7segc_read,
	write:    led_7segc_write,
	open:     led_7segc_open,
	release:  led_7segc_release,
};

static int __init led_7segc_init_module(void)
{
	int ret;

	pr_info(DEV_NAME" ("DEV_VER"): 7seg-LED Driver of SUZAKU I/O Board -LED/SW- for CGI demo.\n");

	sema_init(&sem, 1);
	led_addr = ioremap_nocache(SIL_7SEG_CGI_BASEADDR, SIL_7SEG_CGI_REMAP_SIZE);
	if (led_addr == NULL) {
		printk(KERN_ERR "ioremap failed\n");
		return -EIO;
	}
	ret = register_chrdev(DEV_MAJOR, DEV_NAME, &led_7segc_fops);
	if (ret < 0)
		iounmap(led_addr);

	return ret;
}

static void __exit led_7segc_cleanup_module(void)
{
	iounmap(led_addr);
	unregister_chrdev(DEV_MAJOR, DEV_NAME);
}

module_init(led_7segc_init_module);
module_exit(led_7segc_cleanup_module);

MODULE_AUTHOR("Tetsuya Ohkawa");
MODULE_LICENSE("GPL v2");
