/**
 * drivers/char/sil-led.c -- SUZAKU's I/O Board -LED/SW- LED Driver
 *
 *  Copyright (C) 2006-2007 Atmark Techno, Inc.
 *
 * Tetsuya OHKAWA <tetsuya@atmark-techno.com>
 * Masahiro Hanada <hanada@atmark-techno.com>
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/spinlock.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>
#include <asm/suzaku_sil.h>

#define DEV_NAME           "silled"
#define DEV_MAJOR          70
#define DEV_VER            "1.0.1"

#define LED_MAX            (4)
#define LED_ON             (1)
#define LED_OFF            (0)
#define LED(no)            (1 << ((no)-1))

struct led_handle {
	unsigned int minor;
};
static struct led_handle led_dev[LED_MAX + 1];
static u32 *led_addr;
static struct semaphore sem;

static void update_mem(unsigned long leds, unsigned long cmd)
{
	u32 reg_value;

	reg_value = __raw_readl(led_addr);

	if (cmd == LED_ON) {
		reg_value |=  leds;
	} else if (cmd == LED_OFF) {
		reg_value &= ~leds;
	}

	__raw_writel(reg_value, led_addr);
}

static ssize_t led_read(struct file *filp, char __user *buf,
			size_t count, loff_t *offp)
{
	struct led_handle *dev = filp->private_data;
	ssize_t retval = 0;
	u32 reg_value;
	unsigned char value[2];
	unsigned long cnt = 1;

	if (down_interruptible(&sem)) {
		retval = -ERESTARTSYS;
		goto err;
	}

	if (count == 0)
		goto err;

	if (buf == NULL) {
		retval = -EFAULT;
		goto err;
	}

	reg_value = __raw_readl(led_addr);
	if (dev->minor > 0) {
		reg_value = (reg_value >> (dev->minor - 1)) & 0x1;
	}

	sprintf(value, "%x", reg_value);
	if (copy_to_user(buf, value, cnt)) {
		retval = -EFAULT;
		goto err;
	}

	retval = cnt;
err:
	up(&sem);
	return retval;
}

static ssize_t led_write(struct file *filp, const char __user *buf,
			 size_t count, loff_t *offp)
{
	struct led_handle *dev = filp->private_data;
	ssize_t retval = count;
	char *valbuf = NULL, *endp = NULL;
	unsigned long inval, cmd;

	if (down_interruptible(&sem)) {
		retval = -ERESTARTSYS;
		goto err1;
	}

	if (count == 0)
		goto err1;

	if (buf == NULL) {
		retval = -EFAULT;
		goto err1;
	}

	valbuf = (char *) kmalloc(sizeof(char) * count + 1, GFP_KERNEL);
	if (valbuf == NULL) {
		retval = -EFAULT;
		goto err1;
	}
	if (copy_from_user(valbuf, buf, count)) {
		retval = -EFAULT;
		goto err2;
	}
	valbuf[count] = '\0';

	if (dev->minor == 0) {
		inval = simple_strtoul(valbuf, &endp, 16);
		if (*valbuf == *endp || inval > 0xf) {
			retval = -EINVAL;
			goto err2;
		} 
		__raw_writel(inval, led_addr);

	} else {
		inval = simple_strtoul(valbuf, &endp, 2);
		if (*valbuf == *endp || inval > 0x1) {
			retval = -EINVAL;
			goto err2;
		}

		if (inval == 0x0) {
			cmd = LED_OFF;
		} else /* if (inval == 0x1) */ {
			cmd = LED_ON;
		}
		update_mem(LED(dev->minor), cmd);
	}

err2:
	kfree(valbuf);
err1:
	up(&sem);
	return retval;
}

static int led_open(struct inode *inode, struct file *filp)
{
	int minor;

	minor = MINOR(inode->i_rdev);
	if (minor > LED_MAX) {
		printk(KERN_ERR "illegal minor number\n");
		return -1;
	}

	led_dev[minor].minor = minor;
	filp->private_data = &led_dev[minor];

	return 0;
}

static int led_close(struct inode *inode, struct file *filp)
{
	return 0;
}

static struct file_operations led_fops = {
	owner:   THIS_MODULE,
	open:    led_open,
	release: led_close,
	read:    led_read,
	write:   led_write,
};

static int __init led_init_module(void)
{
	int ret;

	pr_info(DEV_NAME" ("DEV_VER") : LED Driver of SUZAKU I/O Board -LED/SW-\n");

	sema_init(&sem, 1);

	led_addr = (u32 *)ioremap_nocache(SIL_LED_BASEADDR, SIL_LED_REMAP_SIZE);
	if (led_addr == NULL) {
		printk(KERN_ERR "ioremap failed\n");
		return -EIO;
	}
	ret = register_chrdev(DEV_MAJOR, DEV_NAME, &led_fops);
	if (ret < 0) {
		iounmap(led_addr);
		return ret;
	}

	return ret;
}

static void __exit led_cleanup_module(void)
{
	iounmap(led_addr);
	unregister_chrdev(DEV_MAJOR, DEV_NAME);
}

module_init(led_init_module);
module_exit(led_cleanup_module);

MODULE_AUTHOR("Tetsuya Ohkawa");
MODULE_LICENSE("GPL v2");
