/**
 * drivers/char/sil-rsw.c -- SUZAKU's I/O Board -LED/SW- Rotary code switch Driver
 *
 *  Copyright (C) 2006-2007 Atmark Techno, Inc.
 *
 * Tetsuya OHKAWA <tetsuya@atmark-techno.com>
 * Masahiro Hanada <hanada@atmark-techno.com>
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/suzaku_sil.h>

#define DEV_NAME           "silrsw"
#define DEV_MAJOR          73
#define DEV_VER            "1.0.0"

#define RSW_MASK           0xFF
#define POLL_INTERVAL      (HZ/100)       /* 1/100 sec */

static struct timer_list rsw_timer;
static wait_queue_head_t rsw_wait;
static u32 *rsw_addr;

static inline void reset_timer(struct timer_list *timer)
{
	del_timer(timer);
	timer->expires = jiffies + POLL_INTERVAL;
	add_timer(timer);
}

static void timeout_switch(unsigned long data)
{
	wake_up_interruptible(&rsw_wait);
}

static ssize_t rsw_read(struct file *filp, char __user *buf,
			size_t count, loff_t *f_pos)
{
	ssize_t cnt = 1;
	ssize_t retval = 0;
	u32 status;
	char value[2];

	if (count == 0)
		goto err;

	if (buf == NULL) {
		retval = -EFAULT;
		goto err;
	}

	reset_timer(&rsw_timer);
	interruptible_sleep_on(&rsw_wait);

	status = __raw_readl(rsw_addr) & RSW_MASK;
	sprintf(value, "%x", status);
	if (copy_to_user(buf, value, cnt)) {
		retval = -EFAULT;
		goto err;
	}

	retval = cnt;

err:
	return retval;
}

static int rsw_open(struct inode *inode, struct file *filp)
{
	return 0;
}

static int rsw_release(struct inode *inode, struct file *filp)
{
	return 0;
}

struct file_operations rsw_fops = {
	owner:   THIS_MODULE,
	open:    rsw_open,
	read:    rsw_read,
	release: rsw_release,
};

static int __init rsw_init_module(void)
{
	int ret;

	pr_info(DEV_NAME" ("DEV_VER") : Rotary Switch Driver of SUZAKU I/O Board -LED/SW-\n");

	rsw_addr = (u32 *)ioremap_nocache(SIL_RSW_BASEADDR, SIL_RSW_REMAP_SIZE);
	if (rsw_addr == NULL) {
		printk("ioremap failed\n");
		return -EIO;
	}
	ret = register_chrdev(DEV_MAJOR, DEV_NAME, &rsw_fops);
	if (ret < 0) {
		iounmap(rsw_addr);
		return ret;
	}
	init_waitqueue_head(&rsw_wait);
	init_timer(&rsw_timer);
	rsw_timer.function = timeout_switch;

	return ret;
}

static void __exit rsw_cleanup_module(void)
{
	del_timer(&rsw_timer);
	iounmap(rsw_addr);
	unregister_chrdev(DEV_MAJOR, DEV_NAME);
}

module_init(rsw_init_module);
module_exit(rsw_cleanup_module);

MODULE_AUTHOR("Tetsuya Ohkawa");
MODULE_LICENSE("GPL v2");
