/*
 *  linux/drivers/i2c/i2c-suzaku.c
 *
 *  Copyright (C) 2008 Atmark Techno, Inc.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/i2c.h>
#include <linux/i2c-algo-bit.h>
#include <linux/platform_device.h>

#include <asm/io.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#include "../chips/adv71xx.h"

#define AUTHOR "Atmark Techno, Inc."
#define DESCRIPTION "i2c SUZAKU driver"
#define LICENSE "GPL v2"

#define I2C_HW_B_SUZAKU 0x20

#define CHRDEV  "i2c-suzaku"
#define VERSION CHRDEV ": " DESCRIPTION ", (C) 2008 " AUTHOR

static void __iomem *i2c_suzaku_addr = NULL;
#define GPIO_DAT (i2c_suzaku_addr + 0x00)
#define GPIO_DIR (i2c_suzaku_addr + 0x04)

static struct i2c_client *adv7171_client = NULL;
static struct i2c_client *adv7180_client = NULL;

int adv7171_read(unsigned char reg)
{
	if (adv7171_client)
		return adv7171_client->driver->command(adv7171_client, reg,
						       (void *)-1);

	return -EINVAL;
}
EXPORT_SYMBOL(adv7171_read);

int adv7171_write(unsigned char reg, unsigned char val)
{
	if (adv7171_client)
		return adv7171_client->driver->command(adv7171_client, reg,
						       (void *)(int)val);

	return -EINVAL;
}
EXPORT_SYMBOL(adv7171_write);

int adv7180_read(unsigned char reg)
{
	if (adv7180_client)
		return adv7180_client->driver->command(adv7180_client, reg,
						       (void *)-1);

	return -EINVAL;
}
EXPORT_SYMBOL(adv7180_read);

int adv7180_write(unsigned char reg, unsigned char val)
{
	if (adv7180_client)
		return adv7180_client->driver->command(adv7180_client, reg,
						       (void *)(int)val);

	return -EINVAL;
}
EXPORT_SYMBOL(adv7180_write);

static void suzaku_setscl(void *data, int state)
{
	if (state)
		out_be32(GPIO_DAT, in_be32(GPIO_DAT) | 0x2);
	else
		out_be32(GPIO_DAT, in_be32(GPIO_DAT) & ~0x2);
}

static void suzaku_setsda(void *data, int state)
{
	if (state)
		out_be32(GPIO_DIR, in_be32(GPIO_DIR) | 0x1);
	else {
		out_be32(GPIO_DAT, in_be32(GPIO_DAT) & ~0x1);
		out_be32(GPIO_DIR, in_be32(GPIO_DIR) & ~0x1);
	}
}

static int suzaku_getscl(void *data)
{
	return (in_be32(GPIO_DAT) >> 1) & 1;
}

static int suzaku_getsda(void *data)
{
	return (in_be32(GPIO_DAT) >> 0) & 1;
}

static struct i2c_algo_bit_data suzaku_data = {
	.setsda		= suzaku_setsda,
	.setscl		= suzaku_setscl,
	.getsda		= suzaku_getsda,
	.getscl		= suzaku_getscl,
	.udelay		= 20,
	.mdelay		= 20,
	.timeout	= 100
};

static int suzaku_client_reg(struct i2c_client *client)
{
	switch (client->driver->id) {
	case I2C_DRIVERID_ADV7171:
		if (client->addr == I2C_ADV7171_ADDRESS && !adv7171_client)
			adv7171_client = client;
		break;
	case I2C_DRIVERID_ADV7180:
		if (client->addr == I2C_ADV7180_ADDRESS && !adv7180_client)
			adv7180_client = client;
		break;
	}

	return 0;
}

static int suzaku_client_unreg(struct i2c_client *client)
{
	if (client == adv7171_client && client->addr == I2C_ADV7171_ADDRESS)
		adv7171_client = NULL;
	if (client == adv7180_client && client->addr == I2C_ADV7180_ADDRESS)
		adv7180_client = NULL;

	return 0;
}

static struct i2c_adapter suzaku_ops = {
	.name			= "SUZAKU",
	.id			= I2C_HW_B_SUZAKU,
	.algo_data		= &suzaku_data,
	.client_register	= suzaku_client_reg,
	.client_unregister	= suzaku_client_unreg,
};

static int __devinit i2c_suzaku_probe(struct platform_device *pdev)
{
	struct resource *res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	char buf;
	int  ret;

	if (!res) {
		dev_err(&pdev->dev, "no mem for device\n");
		return -ENOENT;
	}

	if (!request_mem_region(res->start, res->end - res->start + 1,
				"i2c-suzaku"))
		return -EBUSY;
	i2c_suzaku_addr = (void __iomem *)
	  ioremap_nocache(res->start, res->end - res->start + 1);
	if (!i2c_suzaku_addr) {
		ret = -EBUSY;
		goto release_mem;
	}

	buf = 0;
	out_be32(GPIO_DAT, in_be32(GPIO_DAT) | 0x2);
	out_be32(GPIO_DIR, (in_be32(GPIO_DIR) & ~0x2) | 0x1);

	ret = i2c_bit_add_bus(&suzaku_ops);
	if (ret < 0) {
		printk (KERN_WARNING CHRDEV ": Unable to add %s bus.\n",
			suzaku_ops.name);
		goto unmap_io;
	}

	printk (KERN_INFO VERSION "\n");
	return 0;

unmap_io:
	iounmap(i2c_suzaku_addr);
release_mem:
	release_mem_region(res->start, res->end - res->start + 1);
	return ret;
}

static int __devexit i2c_suzaku_remove(struct platform_device *pdev)
{
	struct resource *res = platform_get_resource(pdev, IORESOURCE_MEM, 0);

	i2c_bit_del_bus(&suzaku_ops);

	iounmap(i2c_suzaku_addr);
	release_mem_region(res->start, res->end - res->start + 1);

	return 0;
}

static struct platform_driver i2c_suzaku_driver = {
	.probe	= i2c_suzaku_probe,
	.remove	= __devexit_p(i2c_suzaku_remove),
	.driver	= {
		.name	= "i2c-suzaku",
		.owner	= THIS_MODULE,
	},
};

int __init i2c_suzaku_init(void)
{
	static int ret = -1;
	if (!ret)
		return 0;

	ret = platform_driver_register(&i2c_suzaku_driver);
	return ret;
}

void __exit i2c_suzaku_cleanup(void)
{
	platform_driver_unregister(&i2c_suzaku_driver);
}

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESCRIPTION);
MODULE_LICENSE(LICENSE);

module_init(i2c_suzaku_init);
module_exit(i2c_suzaku_cleanup);
