
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/platform_device.h>
#include <linux/slab.h>
#include <linux/i2c.h>
#include <linux/24xx01.h>
#include <linux/delay.h>

#define DRIVER_NAME "24xx01"
#define REVISION    "Rev.1 (2007/--/--)"
#define AUTHOR      "Atmark Techno, Inc."
#define DESCRIPTION "Microchip 24xx01 I2C Serial EEPROM driver"

#define DEBUG
#if defined(DEBUG)
#define DEBUG_FUNC()         printk(DRIVER_NAME ": %s()\n", __FUNCTION__)
#define DEBUG_INFO(args...)  printk(DRIVER_NAME ": " args)
#else
#define DEBUG_FUNC()
#define DEBUG_INFO(args...)
#endif
#define PRINT_INFO(args...)  printk(KERN_INFO  DRIVER_NAME ": " args)
#define PRINT_DEBUG(args...) printk(KERN_DEBUG DRIVER_NAME ": " args)
#define PRINT_ERR(args...)   printk(KERN_ERR   DRIVER_NAME ": " args)

#define I2C_24XX01_ADDR (0x50)
static unsigned short ignore[] = { I2C_CLIENT_END };
static unsigned short normal_addr[] = { I2C_24XX01_ADDR, I2C_CLIENT_END };

static struct i2c_client_address_data addr_data = {
	.normal_i2c	= normal_addr,
	.probe		= ignore,
	.ignore		= ignore,
};

static struct i2c_driver i2c_24xx01_driver;
static struct i2c_client *client_24xx01 = NULL;

int
i2c_24xx01_getdata(u8 addr, u8 *data, u8 len)
{
	int ret;
	u8 buf[1];
	struct i2c_msg msgs[2];

	if (!client_24xx01)
		return -1;

	msgs[0].addr  = client_24xx01->addr;
	msgs[0].flags = 0;
	msgs[0].len   = 1;
	msgs[0].buf   = buf;

	msgs[1].addr  = client_24xx01->addr;
	msgs[1].flags = I2C_M_RD;
	msgs[1].len   = len;
	msgs[1].buf   = data;
	
	buf[0] = addr;
	
	ret = i2c_transfer(client_24xx01->adapter, msgs, 2);
	if (ret != 2) {
		return -1;
	}
	
	return 0;
}
EXPORT_SYMBOL(i2c_24xx01_getdata);

int
i2c_24xx01_setdata(u8 addr, u8 *data, u8 len)
{
	int ret;
	u8 buf[9];
	struct i2c_msg msgs[1];
	u8 count = 0;

	if (!client_24xx01)
		return -1;

	while (len - count){
		int tlen;
		if (len - count >= 8)
			tlen = 8;
		else
			tlen = len - count;

		msgs[0].addr  = client_24xx01->addr;
		msgs[0].flags = 0;
		msgs[0].len   = tlen + 1;
		msgs[0].buf   = buf;
    
		buf[0] = addr + count;
		memcpy(&buf[1], &data[count], tlen);
		
		ret = i2c_transfer(client_24xx01->adapter, msgs, 1);
		if (ret != 1) {
			return -1;
		}
		count += tlen;
		mdelay(5);
	}
	
	return 0;
}
EXPORT_SYMBOL(i2c_24xx01_setdata);

#if defined(DEBUG)
static void
i2c_24xx01_test(void)
{
	int i;
	u8 data[16];
	for (i=0; i<16; i++) {
		data[i] = 0x0 + i;
	}
	i2c_24xx01_setdata(0, data, 16);
	for (i=0; i<16; i++){
		data[i] = 0;
	}
	i2c_24xx01_getdata(0, data, 16);
	for(i=0; i<16; i++){
		DEBUG_INFO("GetData[%d] = %02x\n", i, data[i]);
	}
}
#else
#define i2c_24xx01_test()
#endif

static int
i2c_24xx01_probe(struct i2c_adapter *adap, int addr, int kind)
{
	int ret;

	DEBUG_FUNC();

	client_24xx01 = kzalloc(sizeof(struct i2c_client), GFP_KERNEL);
	if (!client_24xx01)
		return -ENOMEM;

	strcpy(client_24xx01->name, DRIVER_NAME);
	client_24xx01->addr = addr;
	client_24xx01->adapter = adap;
	client_24xx01->driver = &i2c_24xx01_driver;

	ret = i2c_attach_client(client_24xx01);
	if (ret) {
		kfree(client_24xx01);
		client_24xx01 = NULL;
		return ret;
	}

	i2c_24xx01_test();

	return 0;
}

static int
i2c_24xx01_attach(struct i2c_adapter *adap)
{
	DEBUG_FUNC();
	return i2c_probe(adap, &addr_data, i2c_24xx01_probe);
}

static int
i2c_24xx01_detach(struct i2c_client *client)
{
	int ret;
	
	DEBUG_FUNC();
	
	ret = i2c_detach_client(client);
	if (ret)
		return ret;
	
	kfree(client);
	client_24xx01 = NULL;
	
	return 0;
}

#if 0 /* unuse */
static int
i2c_24xx01_command(struct i2c_client *client, unsigned int cmd, void *arg)
{
	DEBUG_FUNC();
	return 0;
}
#endif

static struct i2c_driver i2c_24xx01_driver = {
	.driver	= {
		.name	= DRIVER_NAME,
	},
	.id		= I2C_DRIVERID_24XX01,
	.attach_adapter	= i2c_24xx01_attach,
	.detach_client	= i2c_24xx01_detach,
	//.command	= i2c_24xx01_command,
};

static int __init 
i2c_24xx01_init(void)
{
	int ret;

	DEBUG_FUNC();

	ret = i2c_add_driver(&i2c_24xx01_driver);
	if (ret) {
		PRINT_ERR("Unable to register %s driver.\n", DRIVER_NAME);
		return ret;
	}

	PRINT_INFO(DESCRIPTION " [" REVISION "], (C) 2007 " AUTHOR "\n");

	return 0;
}

static void __exit
i2c_24xx01_exit(void)
{
	DEBUG_FUNC();
	i2c_del_driver(&i2c_24xx01_driver);
}

module_init(i2c_24xx01_init);
module_exit(i2c_24xx01_exit);

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESCRIPTION);
MODULE_LICENSE("GPL");
