/*
 *  adv71xx.c
 *
 *  Copyright (C) 2008 Atmark Techno, Inc.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <asm/io.h>
#include <linux/i2c.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/init.h>
#include <linux/delay.h>

#include "adv71xx.h"

#define AUTHOR "Atmark Techno, Inc."
#define DESCRIPTION "ADV71xx driver"
#define LICENSE "GPL v2"

#define CHRDEV  "adv71xx"
#define VERSION CHRDEV ": ADV71xx driver, (C) 2008 Atmark Techno, Inc."

static struct i2c_driver adv7171_driver;
static struct i2c_driver adv7180_driver;

static unsigned short ignore[] = { I2C_CLIENT_END };
static unsigned short adv7171_normal_addr[] =
  { I2C_ADV7171_ADDRESS, I2C_CLIENT_END };
static unsigned short adv7180_normal_addr[] =
  { I2C_ADV7180_ADDRESS, I2C_CLIENT_END };

static struct i2c_client_address_data adv7171_addr_data = {
	.normal_i2c	= adv7171_normal_addr,
	.probe		= ignore,
	.ignore		= ignore,
};

static struct i2c_client_address_data adv7180_addr_data = {
	.normal_i2c 	= adv7180_normal_addr,
	.probe		= ignore,
	.ignore		= ignore,
};

static int
adv7171_detect(struct i2c_adapter *adap, int addr, int kind)
{
	struct i2c_client *c;
  
	c = kmalloc(sizeof(*c), GFP_KERNEL);
	if (!c)
		return -ENOMEM;

	strcpy(c->name, "ADV7171");
	c->flags   = 0;
	c->addr    = addr;
	c->adapter = adap;
	c->driver  = &adv7171_driver;
	i2c_set_clientdata(c, NULL);
  
	return i2c_attach_client(c);
}

static int
adv7180_detect(struct i2c_adapter *adap, int addr, int kind)
{
	struct i2c_client *c;
  
	c = kmalloc(sizeof(*c), GFP_KERNEL);
	if (!c)
		return -ENOMEM;

	strcpy(c->name, "ADV7180");
	c->flags   = 0;
	c->addr    = addr;
	c->adapter = adap;
	c->driver  = &adv7180_driver;
	i2c_set_clientdata(c, NULL);

	return i2c_attach_client(c);
}

static int
adv7171_attach(struct i2c_adapter *adap)
{
	return i2c_probe(adap, &adv7171_addr_data, adv7171_detect);
}

static int
adv7180_attach(struct i2c_adapter *adap)
{
	return i2c_probe(adap, &adv7180_addr_data, adv7180_detect);
}

static int
adv71xx_detach(struct i2c_client *client)
{
	i2c_detach_client(client);
	kfree(client);
	return 0;
}

static int
adv71xx_read_reg(struct i2c_client *client,
		 unsigned char reg)
{
	int ret;
	unsigned char buf[1];
	struct i2c_msg msgs[2] = {
	  { client->addr, 0, 1, buf },
	  { client->addr, I2C_M_RD, 1, buf }
	};
	buf[0] = reg;

	ret = i2c_transfer(client->adapter, msgs, 2);
	if (ret < 0)
		return ret;
	return buf[0];
}

static int
adv71xx_write_reg(struct i2c_client *client,
		  unsigned char reg, unsigned char data)
{
	int ret;
	unsigned char buf[2];
	struct i2c_msg msgs[1] = {
		{ client->addr, 0, 2, buf }
	};
	buf[0] = reg;
	buf[1] = data;

	ret = i2c_transfer(client->adapter, msgs, 1);
	if (ret < 0)
		return ret;
	return 0;
}

static int
adv71xx_command(struct i2c_client *client, unsigned int cmd, void *arg)
{
	int data = (int)arg;
	if (data < 0)
		return adv71xx_read_reg(client, cmd & 0xff);
	else
		return adv71xx_write_reg(client, cmd & 0xff, data);
}

static struct i2c_driver adv7171_driver = {
	.driver	= {
		.name	= "ADV7171",
	},
	.id		= I2C_DRIVERID_ADV7171,
	.attach_adapter	= adv7171_attach,
	.detach_client	= adv71xx_detach,
	.command	= adv71xx_command,
};

static struct i2c_driver adv7180_driver = {
	.driver = {
		.name	= "ADV7180",
	},
	.id		= I2C_DRIVERID_ADV7180,
	.attach_adapter	= adv7180_attach,
	.detach_client	= adv71xx_detach,
	.command	= adv71xx_command,
};

int __init adv71xx_init(void)
{
	static int ret = -1;
	if (!ret)
		return 0;

	ret = i2c_add_driver(&adv7171_driver);
	if (ret) {
		printk (KERN_WARNING CHRDEV
			": Unable to register %s driver.\n",
			adv7171_driver.driver.name);
		return ret;
	}

	ret = i2c_add_driver(&adv7180_driver);
	if (ret) {
		printk (KERN_WARNING CHRDEV
			": Unable to register %s driver.\n",
			adv7180_driver.driver.name);
		i2c_del_driver(&adv7171_driver);
		return ret;
	}

	printk (KERN_INFO VERSION "\n");
	return 0;
}

void __exit adv71xx_cleanup(void)
{
	i2c_del_driver(&adv7171_driver);
	i2c_del_driver(&adv7180_driver);
}

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESCRIPTION);
MODULE_LICENSE(LICENSE);

module_init(adv71xx_init);
module_exit(adv71xx_cleanup);
