/*
 * drivers/mtd/maps/suzaku.c
 *
 * Copyright (C) 2003-2005, 2007 Atmark Techno, Inc.
 *
 * Yasushi SHOJI <yashi@atmark-techno.com>
 * Masahiro Hanada <hanada@atmark-techno.com>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>

#include <linux/types.h>
#include <linux/ioport.h>

#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>

#include <asm/io.h>

#if defined(CONFIG_PPC)
#define FLASH_START 0xf0000000
#else
#define FLASH_START 0xff000000
#endif
#define FLASH_SIZE  0x00800000
#define FLASH_WIDTH 2
#define FLASH_NAME  "flash"

#define FLASH_4MiB (4*1024*1024)
#define FLASH_8MiB (8*1024*1024)

#define NB_OF(x)  (sizeof(x)/sizeof(x[0]))

static struct mtd_info *flash_mtd;

struct map_info suzaku_flash_map = {
	.name	   = FLASH_NAME,
	.size	   = FLASH_SIZE,
	.phys	   = FLASH_START,
	.bankwidth = FLASH_WIDTH,
};


/*
 * 4MiB Flash:             * 8MiB Flash:              * 8MiB SPI Flash:
 *      0x0 +------------+ *      0x0 +-------------+ *      0x0 +------------+
 *          |    FPGA    | *          | Free1(8KiB) | *          |    FPGA    |
 *  0x80000 +------------+ *  0x10000 +-------------+ * 0x100000 +------------+
 *          |    Boot    | *          | Free2(64KiB)| *          |    Boot    |
 *  0xa0000 +-----+------+ *  0x80000 +-------------+ * 0x120000 +-----+------+
 *          |     | Kern | *          |    FPGA     | *          |     | Kern |
 * 0x210000 | IMG +------+ * 0x100000 +-------------+ * 0x420000 | IMG +------+
 *          |     | User | *          |    Boot     | *          |     | User |
 * 0x3f0000 +-----+------+ * 0x120000 +------+------+ * 0x7f0000 +-----+------+
 *          |   Config   | *          |      | Kern | *          |   Config   |
 *          +------------+ * 0x420000 | IMG  +------| * 0x800000 +------------+
 *                         *          |      | User | *
 *                         * 0x7f0000 +------+------+ *
 *                                    |   Config    |
 *                           0x800000 +-------------+
 */
/*
 * Do _NOT_ change the order of first four entries. These entries are
 * fixed and assinged to fixed minor numbers.
 */
#if defined(CONFIG_SUZAKU_FLASH_4MB) || defined(CONFIG_SUZAKU_FLASH_AUTO)
struct mtd_partition suzaku_partitions_4MiB[] = {
	{ .name = "Flash/All",	      .size = 0x00400000, .offset = 0, },
	{ .name = "Flash/FPGA",	      .size = 0x00080000, .offset = 0, },
	{ .name = "Flash/Bootloader", .size = 0x00020000, .offset = 0x00080000, },
	{ .name = "Flash/Config",     .size = 0x00010000, .offset = 0x003F0000, },
	{ .name = "Flash/Image",      .size = 0x00350000, .offset = 0x000A0000, },
	{ .name = "Flash/Kernel",     .size = 0x00170000, .offset = 0x000A0000, },
	{ .name = "Flash/User",	      .size = 0x001E0000, .offset = 0x00210000, },
};
#endif
#if defined(CONFIG_SUZAKU_FLASH_8MB) || defined(CONFIG_SUZAKU_FLASH_AUTO)
struct mtd_partition suzaku_partitions_8MiB[] = {
	{ .name = "Flash/All",	      .size = 0x00800000, .offset = 0, },
	{ .name = "Flash/FPGA",	      .size = 0x00080000, .offset = 0x00080000, },
	{ .name = "Flash/Bootloader", .size = 0x00020000, .offset = 0x00100000, },
	{ .name = "Flash/Config",     .size = 0x00010000, .offset = 0x007F0000, },
	{ .name = "Flash/Image",      .size = 0x006D0000, .offset = 0x00120000, },
	{ .name = "Flash/Kernel",     .size = 0x00300000, .offset = 0x00120000, },
	{ .name = "Flash/User",	      .size = 0x003D0000, .offset = 0x00420000, },
	{ .name = "Flash/Free1",      .size = 0x00010000, .offset = 0, },
	{ .name = "Flash/Free2",      .size = 0x00070000, .offset = 0x00010000, },
};
#endif
#if defined(CONFIG_MTD_SPI)
struct mtd_partition suzaku_partitions_spi[] = {
        { .name = "Flash/All",        .size = 0x00800000, .offset = 0, },
        { .name = "Flash/FPGA",       .size = 0x00100000, .offset = 0x00000000, },
        { .name = "Flash/Bootloader", .size = 0x00020000, .offset = 0x00100000, },
        { .name = "Flash/Config",     .size = 0x00010000, .offset = 0x007F0000, },
        { .name = "Flash/Image",      .size = 0x006D0000, .offset = 0x00120000, },
        { .name = "Flash/Kernel",     .size = 0x00300000, .offset = 0x00120000, },
        { .name = "Flash/User",       .size = 0x003D0000, .offset = 0x00420000, },
};
#endif

struct resource suzaku_flash_resource = {
	.name  = FLASH_NAME,
	.start = FLASH_START,
	.end   = FLASH_START + FLASH_SIZE - 1,
	.flags = IORESOURCE_IO | IORESOURCE_BUSY,
};

static inline unsigned int is_compatible_mode(void)
{
	return __raw_readl(suzaku_sysreg) & SUZAKU_SYSREG_COMPATIBLE_MODE;
}

static inline unsigned int is_flash_spi(void)
{
	return __raw_readl(suzaku_sysreg) & SUZAKU_SYSREG_FLASH_SPI;
}

static int __init init_suzaku_flash(void)
{
	struct mtd_partition *parts;
	int nb_parts;
	int err;
	char * part_type;

	if (request_resource (&ioport_resource, &suzaku_flash_resource)) {
		printk(KERN_NOTICE "Failed to reserve Suzaku Flash space\n");
		err = -EBUSY;
		goto out;
	}

	suzaku_flash_map.virt = ioremap(FLASH_START, FLASH_SIZE);
	if (!suzaku_flash_map.virt) {
		printk(KERN_NOTICE "Failed to ioremap Suzaku Flash space\n");
		err = -EIO;
		goto out_resource;
	}


	if (is_flash_spi()) {
#ifdef CONFIG_MTD_SPI
		flash_mtd = do_map_probe("spi_probe", &suzaku_flash_map);
#endif
	} else {
#ifdef CONFIG_MTD_CFI
		flash_mtd = do_map_probe("cfi_probe", &suzaku_flash_map);
#endif
	}
	if (!flash_mtd) {
		printk("FLASH probe failed\n");
		err = -ENXIO;
		goto out_ioremap;
	}

	flash_mtd->owner = THIS_MODULE;

	if (is_flash_spi()) {
#ifdef CONFIG_MTD_SPI
		parts = suzaku_partitions_spi;
		nb_parts = NB_OF(suzaku_partitions_spi);
		part_type = "spi";
#endif
	} else {
#if defined(CONFIG_SUZAKU_FLASH_AUTO)
		if (flash_mtd->size == FLASH_8MiB) {
			if (is_compatible_mode()) {
				parts = suzaku_partitions_4MiB;
				nb_parts = ARRAY_SIZE(suzaku_partitions_4MiB);
			} else {
				parts = suzaku_partitions_8MiB;
				nb_parts = ARRAY_SIZE(suzaku_partitions_8MiB);
			}
			part_type = "auto 8MiB";
		} else {
			parts = suzaku_partitions_4MiB;
			nb_parts = ARRAY_SIZE(suzaku_partitions_4MiB);
			part_type = "auto 4MiB";
		}

#elif defined(CONFIG_SUZAKU_FLASH_4MB)
		parts = suzaku_partitions_4MiB;
		nb_parts = ARRAY_SIZE(suzaku_partitions_4MiB);
		part_type = "static 4MiB";

#elif defined(CONFIG_SUZAKU_FLASH_8MB)
		parts = suzaku_partitions_8MiB;
		nb_parts = ARRAY_SIZE(suzaku_partitions_8MiB);
		part_type = "static 8MiB";
#endif
	}

	if (add_mtd_partitions(flash_mtd, parts, nb_parts)) {
		printk("FLASH partitions addition failed\n");
		err = -ENOMEM;
		goto out_probe;
	}

	printk("FLASH partition type: %s %s\n", part_type,
	       is_compatible_mode()?"(4MiB Compatible Mode)":"");

	return 0;

out_probe:
	map_destroy(flash_mtd);
	flash_mtd = NULL;
out_ioremap:
	iounmap(suzaku_flash_map.virt);
out_resource:
	release_resource (&suzaku_flash_resource);
out:
	return err;
}

static int __init init_suzaku_maps(void)
{
	printk(KERN_INFO "Suzaku MTD mappings:\n  Flash 0x%x at 0x%x\n",
	       FLASH_SIZE, FLASH_START);

	init_suzaku_flash();

	return 0;
}

static void __exit cleanup_suzaku_maps(void)
{
	if (flash_mtd) {
		del_mtd_partitions(flash_mtd);
		map_destroy(flash_mtd);
		iounmap(suzaku_flash_map.virt);
		release_resource (&suzaku_flash_resource);
	}
}

module_init(init_suzaku_maps);
module_exit(cleanup_suzaku_maps);

MODULE_AUTHOR("Atmark Techno, Inc.");
MODULE_DESCRIPTION("Suzaku Flash map driver");
MODULE_LICENSE("GPL v2");
