/*
 * siv.c
 *
 * SUZAKU SIV driver
 *
 * 2007-2008 (c) Atmark Techno, Inc.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/ioctl.h>
#include <linux/uaccess.h>
#include <linux/platform_device.h>

#include <asm/io.h>

#include <asm/suzaku_siv.h>
#include "../i2c/chips/adv71xx.h"

#define DRIVER_NAME	"siv"
#define DRIVER_VERSION  "v1.0.0"
#define DRIVER_DESC     "SUZAKU I/O AV Board Video Driver"

struct siv_info {
	int			id;

	struct fb_info		*fb;
	struct device		*dev;

	/* raw memory addresses */
	dma_addr_t		map_dma;	/* physical */
	u_char *		map_cpu;	/* virtual */
	u_int			map_size;

	u32			pseudo_pal[16];
};

MODULE_AUTHOR("Atmark Techno, Inc. <hanada@atmark-techno.com>");
MODULE_DESCRIPTION("Suzaku frame buffer driver");
MODULE_LICENSE("GPL");

static void __iomem *siv_addr = NULL;
#define REG_SIV_CONTROL			(siv_addr + 0x00)
#define     SIV_CONTROL_CORE_RST	0x00000008
#define     SIV_CONTROL_BOARD_RST	0x00000004
#define     SIV_CONTROL_CAPTURE		0x00000002
#define     SIV_CONTROL_DISPLAY		0x00000001
#define REG_SIV_STATUS			(siv_addr + 0x10)
#define REG_SIV_STATUS_CLEAR		(siv_addr + 0x14)
#define     SIV_STATUS_CAP_DONE		0x00000001
#define REG_SIV_INTERRUPT_ENABLE	(siv_addr + 0x40)
#define     SIV_INTERRUPT_ENABLE	0x00000001
#define     SIV_INTERRUPT_DISABLE	0x00000000
#define REG_SIV_EFFECT			(siv_addr + 0x60)
#define     SIV_EFFECT_COLOR		0x00000000
#define     SIV_EFFECT_MONO		0x00000001
#define     SIV_EFFECT_GRAY		0x00000002
#define     SIV_EFFECT_EDGE		0x00000003
#define REG_SIV_CAPTURE_MODE		(siv_addr + 0x64)
#define     SIV_CAPTURE_MODE_CONT	0x00000001
#define     SIV_CAPTURE_MODE_STOP	0x00000000
#define REG_SIV_CAPTURE_ADDRESS		(siv_addr + 0x80)
#define REG_SIV_DISPLAY_ADDRESS		(siv_addr + 0x84)
#define REG_SIV_CORE_ID			(siv_addr + 0xFC)

#define BYTES_PER_PIXEL	4
#define BITS_PER_PIXEL	(BYTES_PER_PIXEL * 8)
#define XRES		720
#define YRES		486
#define XRES_VIRTUAL	XRES
#define YRES_VIRTUAL	YRES
#define LINE_LENGTH	(XRES_VIRTUAL * BYTES_PER_PIXEL)
#define FB_SIZE		(YRES_VIRTUAL * LINE_LENGTH)
#define FBMEM_SIZE	PAGE_ALIGN(FB_SIZE)
#define FB_NUM		3

static struct siv_info *sivinfo[FB_NUM];
static u_char fbmem[FB_NUM][FBMEM_SIZE]
	__attribute__((__aligned__(PAGE_SIZE)));

static DECLARE_WAIT_QUEUE_HEAD(siv_wait);

struct id_struct {
	u16 id;
	u16 version;
	u16 version_mask;
};

static struct id_struct known_boards[] __devinitdata = {
	{ .id = 0x5007, .version = 0x1000, .version_mask = 0xfff0 },
	{ },
};

static int __devinit is_known_board(u16 id, u16 version)
{
	int ret = 0;
	int i;

	for (i=0; known_boards[i].id; i++) {
		if (known_boards[i].id == id) {
			if (known_boards[i].version ==
			    (version & known_boards[i].version_mask)) {
				ret = 1;
				break;
			}
		}
	}

	return ret;
}

static int
siv_check_var(struct fb_var_screeninfo *var,
	      struct fb_info *info)
{
	/* validate x/y resolution */

	var->xres = XRES;
	var->yres = YRES;
	var->xres_virtual = XRES_VIRTUAL;
	var->yres_virtual = YRES_VIRTUAL;
	var->xoffset = 0;
	var->yoffset = 0;

	/* validate bpp */

	var->bits_per_pixel = BITS_PER_PIXEL;

	/* set r/g/b positions */

	var->red.offset		= 16;
	var->green.offset	= 8;
	var->blue.offset	= 0;
	var->transp.offset	= 24;
	var->red.length		= 8;
	var->green.length	= 8;
	var->blue.length	= 8;
	var->transp.length	= 8;

	return 0;
}

static int
siv_set_par(struct fb_info *info)
{
	struct siv_info *fbi = info->par;

	fbi->fb->fix.visual = FB_VISUAL_TRUECOLOR;
	fbi->fb->fix.line_length = LINE_LENGTH;

	return 0;
}

static inline unsigned int
chan_to_field(unsigned int chan, struct fb_bitfield *bf)
{
	chan &= 0xffff;
	chan >>= 16 - bf->length;
	return chan << bf->offset;
}

static int
siv_setcolreg(unsigned regno,
	      unsigned red, unsigned green, unsigned blue,
	      unsigned transp, struct fb_info *info)
{
	struct siv_info *fbi = info->par;
	unsigned int val;

	switch (fbi->fb->fix.visual) {
	case FB_VISUAL_TRUECOLOR:
		/* true-colour, use pseuo-palette */

		if (regno < 16) {
			u32 *pal = fbi->fb->pseudo_palette;

			val  = chan_to_field(red,    &fbi->fb->var.red);
			val |= chan_to_field(green,  &fbi->fb->var.green);
			val |= chan_to_field(blue,   &fbi->fb->var.blue);
			val |= chan_to_field(transp, &fbi->fb->var.transp);

			pal[regno] = val;
		}
		break;

	default:
		return 1;   /* unknown type */
	}

	return 0;
}

static int
siv_blank(int blank_mode, struct fb_info *info)
{
	if (blank_mode == FB_BLANK_UNBLANK)
		/* turn on panel */
		out_be32(REG_SIV_CONTROL,
			 in_be32(REG_SIV_CONTROL) | SIV_CONTROL_DISPLAY);
	else {
		/* turn off panel */
		out_be32(REG_SIV_CONTROL,
			 in_be32(REG_SIV_CONTROL) & ~SIV_CONTROL_DISPLAY);
	}

	return 0;
}

static int
siv_ioctl(struct fb_info *info, unsigned int cmd, unsigned long arg)
{
	struct codec_param param;
	int ret;

	switch (cmd) {
	case FBIO_DISPLAY_ENABLE:
		out_be32(REG_SIV_CONTROL,
			 in_be32(REG_SIV_CONTROL) | SIV_CONTROL_DISPLAY);
		return 0;

	case FBIO_DISPLAY_DISABLE:
		out_be32(REG_SIV_CONTROL,
			 in_be32(REG_SIV_CONTROL) & ~SIV_CONTROL_DISPLAY);
		return 0;

	case FBIO_CAPTURE_ENABLE:
		out_be32(REG_SIV_CONTROL,
			 in_be32(REG_SIV_CONTROL) | SIV_CONTROL_CAPTURE);
		return 0;

	case FBIO_CAPTURE_DISABLE:
		out_be32(REG_SIV_CONTROL,
			 in_be32(REG_SIV_CONTROL) & ~SIV_CONTROL_CAPTURE);
		return 0;

	case FBIO_DISPLAY_SELECT:
		out_be32(REG_SIV_DISPLAY_ADDRESS,
			 ((struct siv_info *)info->par)->map_dma);
		return 0;

	case FBIO_CAPTURE_SELECT:
		out_be32(REG_SIV_CAPTURE_ADDRESS,
			 ((struct siv_info *)info->par)->map_dma);
		out_be32(REG_SIV_STATUS_CLEAR, SIV_STATUS_CAP_DONE);
		out_be32(REG_SIV_STATUS_CLEAR, 0);
		out_be32(REG_SIV_INTERRUPT_ENABLE, SIV_INTERRUPT_ENABLE);
		return 0;

	case FBIO_CAPTURE_MODE_SELECT:
		switch (arg) {
		case CAPTURE_CONTINUOUS_MODE:
			out_be32(REG_SIV_CAPTURE_MODE, SIV_CAPTURE_MODE_CONT);
			break;
		case CAPTURE_STOP_MODE:
			out_be32(REG_SIV_CAPTURE_MODE, SIV_CAPTURE_MODE_STOP);
			break;
		}
		return 0;

	case FBIO_EFFECT_SELECT:
		switch (arg) {
		case EFFECT_COLOR:
			out_be32(REG_SIV_EFFECT, SIV_EFFECT_COLOR);
			break;
		case EFFECT_MONO:
			out_be32(REG_SIV_EFFECT, SIV_EFFECT_MONO);
			break;
		case EFFECT_GRAY:
			out_be32(REG_SIV_EFFECT, SIV_EFFECT_GRAY);
			break;
		case EFFECT_EDGE:
			out_be32(REG_SIV_EFFECT, SIV_EFFECT_EDGE);
			break;
		default:
			return -EINVAL;
		}
		return 0;

	case FBIO_WAIT_CAPTURE:
		return wait_event_interruptible
		  (siv_wait, in_be32(REG_SIV_STATUS) & SIV_STATUS_CAP_DONE);

	case FBIO_READ_ADV7171_REG:
		copy_from_user(&param, (void *)arg, sizeof(param));
		ret = adv7171_read(param.subaddr);
		if (ret < 0)
			return ret;
		param.data = ret;
		copy_to_user((void *)arg, &param, sizeof(param));
		return 0;

	case FBIO_WRITE_ADV7171_REG:
		copy_from_user(&param, (void *)arg, sizeof(param));
		return adv7171_write(param.subaddr, param.data);

	case FBIO_READ_ADV7180_REG:
		copy_from_user(&param, (void *)arg, sizeof(param));
		ret = adv7180_read(param.subaddr);
		if (ret < 0)
			return ret;
		param.data = ret;
		copy_to_user((void *)arg, &param, sizeof(param));
		return 0;

	case FBIO_WRITE_ADV7180_REG:
		copy_from_user(&param, (void *)arg, sizeof(param));
		return adv7180_write(param.subaddr, param.data);

	default:
				return -EINVAL;
	}
}

static int
siv_mmap(struct fb_info *info, struct vm_area_struct *vma)
{
	unsigned long off, start;
	u32 len;

	off = vma->vm_pgoff << PAGE_SHIFT;

	/* frame buffer memory */
	start = info->fix.smem_start;
	len = PAGE_ALIGN((start & ~PAGE_MASK) + info->fix.smem_len);
	pgprot_val(vma->vm_page_prot) |= _PAGE_WRITETHRU;
	start &= PAGE_MASK;
	if ((vma->vm_end - vma->vm_start + off) > len)
       		return -EINVAL;
	off += start;
	vma->vm_pgoff = off >> PAGE_SHIFT;
	if (io_remap_pfn_range(vma, vma->vm_start, off >> PAGE_SHIFT,
			       vma->vm_end - vma->vm_start, vma->vm_page_prot))
		return -EAGAIN;

	return 0;
}

static struct fb_ops siv_ops = {
	.owner		= THIS_MODULE,
	.fb_check_var	= siv_check_var,
	.fb_set_par	= siv_set_par,
	.fb_blank	= siv_blank,
	.fb_setcolreg	= siv_setcolreg,
	.fb_fillrect	= cfb_fillrect,
	.fb_copyarea	= cfb_copyarea,
	.fb_imageblit	= cfb_imageblit,
	.fb_ioctl	= siv_ioctl,
	.fb_mmap	= siv_mmap,
};

static int __devinit
siv_map_video_memory(struct siv_info *fbi)
{
	fbi->map_cpu  = fbmem[fbi->id];
	fbi->map_dma  = virt_to_phys(fbi->map_cpu);
	fbi->map_size = fbi->fb->fix.smem_len;

	/* prevent initial garbage on screen */
	memset(fbi->map_cpu, 0x00, fbi->map_size);

	fbi->fb->screen_base	= fbi->map_cpu;
	fbi->fb->fix.smem_start = fbi->map_dma;

	pr_debug("map_video_memory: dma=%08x cpu=%p size=%08x\n",
		 fbi->map_dma, fbi->map_cpu, fbi->fb->fix.smem_len);

	return 0;
}

static inline void
siv_unmap_video_memory(struct siv_info *fbi, int id)
{
        /* Turn off the display and capture. */
        out_be32(REG_SIV_CONTROL, 0);
}

static int __devinit
siv_init_registers(void)
{
	const u8 reg7180[][2] = {
	  { 0x00, 0x05 },	// INSEL=CVBS in on AIN6
	  { 0x04, 0xd5 },	// Disable SFL
	  { 0x17, 0x41 },	// select SH1
	  { 0x31, 0x02 },	// Clears NEWAV_MODE,
				//   SAV/EAV to suit ADV video encoders
	  { 0x3d, 0xa2 },	// *MWE enable Manual Window,
				//   color kill threshold to 2
	  { 0x3e, 0x6a },	// BML optimization
	  { 0x3f, 0xa0 },	// *BGB optimization
	  { 0x0e, 0x80 },	// Hidden space
	  { 0x55, 0x81 },	// ADC configuration
	  { 0x0e, 0x00 },	// User space
	  { 0x37, 0x01 },	// Polarity normal
	  { 0xf4, 0x1f },	// CLK(High Drive 4X) DATA(Midium 2X)
	};
	const u8 reg7171[][2] = {
	  { 0x02, 0x02 },	// Use Sub Calire
	  { 0x09, 0x1f },	// Sub Calire 0
	  { 0x0a, 0x7c },	// Sub Calire 0
	  { 0x0b, 0xf0 },	// Sub Calire 0
	  { 0x0c, 0x21 },	// Sub Calire 0
	};
	int i;
	int ret;

	out_be32(REG_SIV_CONTROL,
		 SIV_CONTROL_CORE_RST | SIV_CONTROL_BOARD_RST);
	mdelay(10);
	out_be32(REG_SIV_CONTROL, 0);
	mdelay(10);

	for (i = 0; i < sizeof(reg7180) / sizeof(reg7180[0]); i++) {
		ret = adv7180_write(reg7180[i][0], reg7180[i][1]);
#ifdef _DEBUG
		pr_debug("adv7180_write(%02x, %02x);\n",
			 reg7180[i][0], reg7180[i][1]);
#endif
		if (ret < 0)
			return ret;
	}
	for (i = 0; i < sizeof(reg7171) / sizeof(reg7171[0]); i++) {
		ret = adv7171_write(reg7171[i][0], reg7171[i][1]);
#ifdef _DEBUG
		pr_debug("adv7171_write(%02x, %02x);\n",
			 reg7171[i][0], reg7171[i][1]);
#endif
		if (ret < 0)
			return ret;
	}

#ifdef _DEBUG
	for (i = 0; i < 256; i++)
		pr_debug("adv7180-%02x: %02x\n", i, adv7180_read(i));
	for (i = 0; i < 26; i++)
		pr_debug("adv7171-%02x: %02x\n", i, adv7171_read(i));
#endif

	out_be32(REG_SIV_DISPLAY_ADDRESS, sivinfo[0]->map_dma);
	out_be32(REG_SIV_CAPTURE_ADDRESS, sivinfo[1]->map_dma);

	out_be32(REG_SIV_EFFECT, SIV_EFFECT_COLOR);
	out_be32(REG_SIV_CAPTURE_MODE, SIV_CAPTURE_MODE_CONT);
	out_be32(REG_SIV_INTERRUPT_ENABLE, SIV_INTERRUPT_DISABLE);
	out_be32(REG_SIV_STATUS_CLEAR, SIV_STATUS_CAP_DONE);
	out_be32(REG_SIV_STATUS_CLEAR, 0);

        /* Turn on the display and capture. */
        out_be32(REG_SIV_CONTROL, SIV_CONTROL_DISPLAY | SIV_CONTROL_CAPTURE);

	return 0;
}

static irqreturn_t
siv_irq(int irq, void *dev_id, struct pt_regs *r)
{
	if (waitqueue_active(&siv_wait))
		wake_up(&siv_wait);

	out_be32(REG_SIV_INTERRUPT_ENABLE, SIV_INTERRUPT_DISABLE);

	return IRQ_HANDLED;
}

static char driver_name[] = DRIVER_NAME;

static int __devinit
siv_register_fb(struct platform_device *pdev, int id)
{
	struct siv_info *info;
	struct fb_info *fbinfo;
	int ret;

	fbinfo = framebuffer_alloc(sizeof(struct siv_info), &pdev->dev);
	if (!fbinfo)
		return -ENOMEM;

	info = fbinfo->par;
	info->id = id;
	info->fb = fbinfo;
	info->dev = &pdev->dev;
	sivinfo[id] = info;

	strcpy(fbinfo->fix.id, driver_name);

	fbinfo->fix.type	    = FB_TYPE_PACKED_PIXELS;
	fbinfo->fix.type_aux	    = 0;
	fbinfo->fix.visual	    = FB_VISUAL_TRUECOLOR;
	fbinfo->fix.line_length	    = LINE_LENGTH;
	fbinfo->fix.xpanstep	    = 0;
	fbinfo->fix.ypanstep	    = 0;
	fbinfo->fix.ywrapstep	    = 0;
	fbinfo->fix.accel	    = FB_ACCEL_NONE;

	fbinfo->var.nonstd	    = 0;
	fbinfo->var.activate	    = FB_ACTIVATE_NOW;
	fbinfo->var.height	    = XRES;
	fbinfo->var.width	    = YRES;
	fbinfo->var.accel_flags     = 0;
	fbinfo->var.vmode	    = FB_VMODE_NONINTERLACED;

	fbinfo->fbops		    = &siv_ops;
	fbinfo->flags		    = FBINFO_FLAG_DEFAULT;
	fbinfo->pseudo_palette      = &info->pseudo_pal;

	fbinfo->var.xres	    = XRES;
	fbinfo->var.xres_virtual    = XRES_VIRTUAL;
	fbinfo->var.yres	    = YRES;
	fbinfo->var.yres_virtual    = YRES_VIRTUAL;
	fbinfo->var.bits_per_pixel  = BITS_PER_PIXEL;

	fbinfo->var.red.offset      = 16;
	fbinfo->var.green.offset    = 8;
	fbinfo->var.blue.offset     = 0;
	fbinfo->var.transp.offset   = 24;
	fbinfo->var.red.length      = 8;
	fbinfo->var.green.length    = 8;
	fbinfo->var.blue.length     = 8;
	fbinfo->var.transp.length   = 8;
	fbinfo->fix.smem_len        = FB_SIZE;

	fb_alloc_cmap(&fbinfo->cmap, 16, 0);

	/* Initialize video memory */
	ret = siv_map_video_memory(info);
	if (ret) {
		dev_err(info->dev, "Failed to allocate video RAM: %d\n", ret);
		ret = -ENOMEM;
		goto dealloc_fb;
	}

	ret = siv_check_var(&fbinfo->var, fbinfo);

	ret = register_framebuffer(fbinfo);
	if (ret < 0) {
		dev_err(info->dev,
			"Failed to register framebuffer device: %d\n", ret);
		goto free_video_memory;
	}

	pr_info("fb%d: %s frame buffer device\n",
	       fbinfo->node, fbinfo->fix.id);

	return 0;

free_video_memory:
	siv_unmap_video_memory(info, id);
dealloc_fb:
	fb_dealloc_cmap(&fbinfo->cmap);
	framebuffer_release(fbinfo);
	return ret;
}

static int __devinit
siv_probe(struct platform_device *pdev)
{
	int irq = platform_get_irq(pdev, 0);
	struct resource *res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	u32 tmp;
	u16 board_id, board_version;
	int ret;
	int i;

	if (irq < 0) {
		dev_err(&pdev->dev, "no irq for device\n");
		return -ENOENT;
	}
	if (!res) {
		dev_err(&pdev->dev, "no mem for device\n");
		return -ENOENT;
	}

	if (!request_mem_region(res->start, res->end - res->start + 1,
				DRIVER_NAME)) {
		return -EBUSY;
	}
	siv_addr = (void __iomem *)ioremap_nocache(res->start,
						   res->end - res->start + 1);
	if (!siv_addr) {
		ret = -EBUSY;
		goto release_mem;
	}

	tmp = in_be32(REG_SIV_CORE_ID);
	board_id = (tmp >> 16) & 0xffff;
	board_version = tmp & 0xffff;

	pr_info("  OPB-SIV00 (id: %#hx; ver: %#hx)\n",
		board_id, board_version);

	if (!is_known_board(board_id, board_version)) {
		dev_err(&pdev->dev, "  Unsupported ID and/or Version\n");
		ret = -ENODEV;
		goto unmap_io;
	}

	ret = request_irq(irq, siv_irq, IRQF_DISABLED, pdev->name, NULL);
	if (ret) {
		dev_err(&pdev->dev, "cannot get irq %d - err %d\n", irq, ret);
		ret = -EBUSY;
		goto unmap_io;
	}

	for (i = 0; i < FB_NUM; i++) {
		ret = siv_register_fb (pdev, i);
		if (ret < 0) {
			for (i--; i >= 0; i--) {
				struct siv_info *info;
				struct fb_info *fbinfo;

				info = sivinfo[i];
				fbinfo = info->fb;
				siv_unmap_video_memory(info, i);
				unregister_framebuffer(fbinfo);
				fb_dealloc_cmap(&fbinfo->cmap);
				framebuffer_release(fbinfo);
			}
			goto release_irq;
		}
	}

	ret = siv_init_registers();
	if (ret < 0)
		goto unregister_fb;

	return 0;

unregister_fb:
	for (i = 0; i < FB_NUM; i++) {
		struct siv_info *info;
		struct fb_info *fbinfo;

		info = sivinfo[i];
		fbinfo = info->fb;
		siv_unmap_video_memory(info, i);
		unregister_framebuffer(fbinfo);
		fb_dealloc_cmap(&fbinfo->cmap);
		framebuffer_release(fbinfo);
	}
release_irq:
	free_irq(irq, NULL);
unmap_io:
	iounmap(siv_addr);
release_mem:
	release_mem_region(res->start, res->end - res->start + 1);
	return ret;
}

static int __devexit
siv_remove(struct platform_device *pdev)
{
	struct fb_info *fbinfo;
	struct siv_info *info;
	int irq = platform_get_irq(pdev, 0);
	struct resource *res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	int i;

	for (i = 0; i < FB_NUM; i++) {
		info = sivinfo[i];
		fbinfo = info->fb;

		siv_unmap_video_memory(info, i);

		unregister_framebuffer(fbinfo);
		fb_dealloc_cmap(&fbinfo->cmap);
		framebuffer_release(fbinfo);
	}

	free_irq(irq, NULL);
	iounmap(siv_addr);
	release_mem_region(res->start, res->end - res->start + 1);

	return 0;
}

static struct platform_driver siv_driver = {
	.probe	= siv_probe,
	.remove	= __devexit_p(siv_remove),
	.driver	= {
		.name	= DRIVER_NAME,
		.owner	= THIS_MODULE,
	},
};

#ifndef MODULE
extern int __init i2c_suzaku_init(void);
extern int __init adv71xx_init(void);
#endif
int __init
siv_init(void)
{
#ifndef MODULE
	int ret;

	ret = i2c_suzaku_init();
	if (ret)
		return ret;

	ret = adv71xx_init();
	if (ret)
		return ret;
#endif

	printk("%s (%s) : %s\n", DRIVER_NAME, DRIVER_VERSION, DRIVER_DESC);

	return platform_driver_register(&siv_driver);
}

void __exit
siv_cleanup(void)
{
	platform_driver_unregister(&siv_driver);
}

module_init(siv_init);
module_exit(siv_cleanup);
