#include <linux/module.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/device.h>
#include <linux/spinlock.h>
#include <linux/interrupt.h>
#include <linux/err.h>
#include <linux/dma-mapping.h>
#include <linux/platform_device.h>

#include <linux/tlv320aic.h>

#include <asm/io.h>
#include <asm/irq.h>

#include <sound/driver.h>
#include <sound/core.h>
#include <sound/initval.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/control.h>

#define DRIVER_NAME	"sia"
#define DRIVER_VERSION  "1.0"
#define DRIVER_DESC     "SUZAKU I/O AV Board Audio Driver"
#define DRIVER_AUTHOR   "Atmark Techno, Inc."

#ifdef CONFIG_SND_DEBUG
#define DEBUG_FUNC()	pr_debug(DRIVER_NAME ": %s()\n", __FUNCTION__)
#else
#define DEBUG_FUNC()
#endif

#define DEFAULT_FORMAT			(SNDRV_PCM_FORMAT_S16_BE)
#define DEFAULT_SAMPLING_RATE		(48000)
#define DEFAULT_LINE_INPUT_VOLUME	(0x1c1c) //  0dB
#define DEFAULT_HEADPHONE_VOLUME	(0x2121) //-40dB

static void __iomem *sia_addr = NULL;
#define REG_SIA_CONTROL			(sia_addr + 0x00)
#define     SIA_CONTROL_BYTE_SWAP	0x00000004
#define     SIA_CONTROL_REC_RST		0x00000002
#define     SIA_CONTROL_PLAY_RST	0x00000001
#define REG_SIA_STATUS			(sia_addr + 0x10)
#define     SIA_RX_FIFO_FULL		0x00000020
#define     SIA_TX_FIFO_EMP		0x00000010
#define REG_SIA_INTERRUPT_ENABLE	(sia_addr + 0x40)
#define     SIA_RX_FIFO_HFULL		0x00000008
#define     SIA_RX_FIFO_VALID		0x00000004
#define     SIA_TX_FIFO_FULL		0x00000002
#define     SIA_TX_FIFO_HEMP		0x00000001
#define REG_SIA_MONO_SELECT		(sia_addr + 0x64)
#define     SIA_STEREO			0x00000000
#define     SIA_MONO			0x00000001
#define REG_SIA_AUDIO_RX_FIFO		(sia_addr + 0x80)
#define REG_SIA_AUDIO_TX_FIFO		(sia_addr + 0x84)
#define REG_SIA_AUDIO_RX_FIFO_DEPTH	(sia_addr + 0xa0)
#define REG_SIA_AUDIO_TX_FIFO_DEPTH	(sia_addr + 0xa4)
#define REG_SIA_CORE_ID			(sia_addr + 0xfc)

struct id_struct {
	u16 id;
	u16 version;
	u16 version_mask;
};

static struct id_struct known_boards[] __devinitdata = {
	{ .id = 0x5008, .version = 0x1000, .version_mask = 0xfff0 },
	{ },
};

static int __devinit is_known_board(u16 id, u16 version)
{
	int i;

	for (i=0; known_boards[i].id; i++) {
		if (known_boards[i].id == id) {
			if (known_boards[i].version ==
			    (version & known_boards[i].version_mask)) {
				return 1;
			}
		}
	}

	return 0;
}

typedef struct sia_runtime {
	u32			control_fifo_reset_bit;
	u32			interrupt_enable_bit;

	snd_pcm_substream_t	*substream;

	/*
	 * PIO support
	 */
	void			*start;
	void			*end;
	void			*ptr;
	int			bytes;
	unsigned int		period;
	unsigned int		fifosz;
} sia_runtime_t;

typedef struct sia {
	struct device		*dev;

	u32			control;
	u32			interrupt_enable;

	spinlock_t		lock;

	sia_runtime_t		playback;
	sia_runtime_t		capture;

	snd_pcm_t		*pcm;
} sia_t;

typedef enum {
	MIXERID_LINE_INPUT_VOL,
	MIXERID_HEADPHONE_VOL,
} volume_type_t;

static void
codec_dump_register(void)
{
	u16 livol, rivol, lhvol, rhvol;
	u16 apath, dpath, pwr, diform, srate, dact;

	livol = tlv320aic_getreg(TLV_LIVOL);
	rivol = tlv320aic_getreg(TLV_RIVOL);
	lhvol = tlv320aic_getreg(TLV_LHVOL);
	rhvol = tlv320aic_getreg(TLV_RHVOL);
	apath = tlv320aic_getreg(TLV_APATH);
	dpath = tlv320aic_getreg(TLV_DPATH);
	pwr = tlv320aic_getreg(TLV_PWR);
	diform = tlv320aic_getreg(TLV_DIFORM);
	srate = tlv320aic_getreg(TLV_SRATE);
	dact = tlv320aic_getreg(TLV_DACT);

	pr_debug("LIVOL : 0x%04x,  RIVOL  : 0x%04x\n", livol, rivol);
	pr_debug("LHVOL : 0x%04x,  RHVOL  : 0x%04x\n", lhvol, rhvol);
	pr_debug("APATH : 0x%04x,  DPATH  : 0x%04x\n", apath, dpath);
	pr_debug("PWR   : 0x%04x,  DIFORM : 0x%04x\n", pwr, diform);
	pr_debug("SRATE : 0x%04x,  DACT   : 0x%04x\n", srate, dact);
}

static void
i2s_dump_register(void)
{
	u32 control, status, ie, mono;

	control	= in_be32(REG_SIA_CONTROL);
	status	= in_be32(REG_SIA_STATUS);
	ie	= in_be32(REG_SIA_INTERRUPT_ENABLE);
	mono	= in_be32(REG_SIA_MONO_SELECT);

	pr_debug("Control         : 0x%08x, Status          : 0x%08x\n",
		 control, status);
	pr_debug("Interrupt Enable: 0x%08x, Monaural Select : 0x%08x\n",
		 ie, mono);
}

static irqreturn_t
sia_irq(int irq, void *devid, struct pt_regs *regs)
{
	sia_t *sia = devid;
	sia_runtime_t *siarun;
	u32 *ptr;
	unsigned int len;
	unsigned int val;

	spin_lock(&sia->lock);

	siarun = &sia->playback;

	if (siarun->substream && siarun->start) {
		ptr = siarun->ptr;
		do {
			len = siarun->fifosz;

			if (siarun->bytes <= 0) {
				siarun->bytes += siarun->period;
				siarun->ptr = ptr;
				spin_unlock(&sia->lock);
				snd_pcm_period_elapsed(siarun->substream);
				spin_lock(&sia->lock);
			}

			val = in_be32(REG_SIA_STATUS);
			if (!(val & SIA_TX_FIFO_HEMP))
				break;
			if (!(val & SIA_TX_FIFO_EMP))
				len >>= 1;

			siarun->bytes -= len;

			for (; len > 0; len -= 4) {
				out_be32(REG_SIA_AUDIO_TX_FIFO, *ptr++);

				if ((void *)ptr >= siarun->end)
					ptr = siarun->start;
			}
		}
		while (1);

		siarun->ptr = ptr;
	}
	else {
		len = siarun->fifosz;

		val = in_be32(REG_SIA_STATUS);
		if (val & SIA_TX_FIFO_HEMP) {
			if (!(val & SIA_TX_FIFO_EMP))
				len >>= 1;

			for (; len > 0; len -= 4)
				out_be32(REG_SIA_AUDIO_TX_FIFO, 0);
		}
	}

	siarun = &sia->capture;

	if (siarun->substream && siarun->start) {
		ptr = siarun->ptr;
		do {
			len = siarun->fifosz;

			if (siarun->bytes <= 0) {
				siarun->bytes += siarun->period;
				siarun->ptr = ptr;
				spin_unlock(&sia->lock);
				snd_pcm_period_elapsed(siarun->substream);
				spin_lock(&sia->lock);
			}

			val = in_be32(REG_SIA_STATUS);
			if (!(val & SIA_RX_FIFO_HFULL))
				break;
			if (!(val & SIA_RX_FIFO_FULL))
				len >>= 1;

			siarun->bytes -= len;

			for (; len > 0; len -= 4) {
				*ptr++ = in_be32(REG_SIA_AUDIO_RX_FIFO);

				if ((void *)ptr >= siarun->end)
					ptr = siarun->start;
			}
		}
		while (1);

		siarun->ptr = ptr;
	}
	else {
		len = siarun->fifosz;

		val = in_be32(REG_SIA_STATUS);
		if (val & SIA_TX_FIFO_HEMP) {
			if (!(val & SIA_TX_FIFO_EMP))
				len >>= 1;

			for (; len > 0; len -= 4)
				in_be32(REG_SIA_AUDIO_RX_FIFO);
		}
	}

	spin_unlock(&sia->lock);

	return IRQ_HANDLED;
}

static int
sia_audio_get_volume_info(sia_t *sia, int type, snd_ctl_elem_info_t *uinfo)
{
	DEBUG_FUNC();

	switch (type) {
	case MIXERID_LINE_INPUT_VOL:
		uinfo->value.integer.min = 0;
		uinfo->value.integer.max = 0x1f;
		uinfo->value.integer.step = 1;
		break;
	case MIXERID_HEADPHONE_VOL:
		uinfo->value.integer.min = 0;
		uinfo->value.integer.max = 0x4f;
		uinfo->value.integer.step = 1;
		break;
	default:
		return -ENODEV;
	}

	uinfo->type = SNDRV_CTL_ELEM_TYPE_INTEGER;
	uinfo->count = 2;

	return 0;
}

static int
sia_audio_get_volume(sia_t *sia, int type)
{
	u8 left, right;
	u8 lval, rval;

	DEBUG_FUNC();

	switch (type) {
	case MIXERID_LINE_INPUT_VOL:
		{
			u16 livol, rivol;
			livol = tlv320aic_getreg(TLV_LIVOL);
			left = (livol & TLV_LINE_INPUT_VOL_MASK);

			rivol = tlv320aic_getreg(TLV_RIVOL);
			right = (livol & TLV_LINE_INPUT_VOL_MASK);
			break;
		}
	case MIXERID_HEADPHONE_VOL:
		{
			u16 lhvol, rhvol;

			lhvol = tlv320aic_getreg(TLV_LHVOL);
			lval = (lhvol & TLV_HEADPHONE_VOL_MASK);
			if (lval < TLV_HEADPHONE_VOL_MIN) {
				lval = TLV_HEADPHONE_VOL_MIN;
				lhvol =
				  (lhvol & ~TLV_HEADPHONE_VOL_MASK) | lval;
				tlv320aic_setreg(TLV_LHVOL, lhvol);
			}
			left = lval - TLV_HEADPHONE_VOL_MIN;

			rhvol = tlv320aic_getreg(TLV_RHVOL);
			rval = (rhvol & TLV_HEADPHONE_VOL_MASK);
			if (rval < TLV_HEADPHONE_VOL_MIN) {
				rval = TLV_HEADPHONE_VOL_MIN;
				rhvol =
				  (rhvol & ~TLV_HEADPHONE_VOL_MASK) | rval;
				tlv320aic_setreg(TLV_RHVOL, rhvol);
			}
			right = rval - TLV_HEADPHONE_VOL_MIN;
		}
		break;
	default:
		return -ENODEV;
	}
	return (left & 0xff) | ((right & 0xff) << 8);
}

static int
sia_audio_set_volume(sia_t *sia, int type, int volume)
{
	u8 left, right;
	u16 lval, rval;
	int change = 0;

	DEBUG_FUNC();

	left  = (volume & 0x00ff);
	right = ((volume & 0xff00) >> 8);

	switch (type) {
	case MIXERID_LINE_INPUT_VOL:
		{
			u16 livol, rivol;

			lval = left;
			if (lval > TLV_LINE_INPUT_VOL_MAX)
				lval = TLV_LINE_INPUT_VOL_MAX;
			livol = tlv320aic_getreg(TLV_LIVOL);
			if (lval != (livol & TLV_LINE_INPUT_VOL_MASK)) {
				livol =
				  (livol & ~TLV_LINE_INPUT_VOL_MASK) | lval;
				tlv320aic_setreg(TLV_LIVOL, livol);
				change = 1;
			}

			rval = right;
			if (rval > TLV_LINE_INPUT_VOL_MAX)
				rval = TLV_LINE_INPUT_VOL_MAX;
			rivol = tlv320aic_getreg(TLV_RIVOL);
			if (rval != (rivol & TLV_LINE_INPUT_VOL_MASK)) {
				rivol =
				  (rivol & ~TLV_LINE_INPUT_VOL_MASK) | rval;
				tlv320aic_setreg(TLV_RIVOL, rivol);
				change = 1;
			}
			break;
		}
	case MIXERID_HEADPHONE_VOL:
		{
			u16 lhvol, rhvol;
			lval = left + TLV_HEADPHONE_VOL_MIN;
			if (lval > TLV_HEADPHONE_VOL_MAX)
				lval = TLV_HEADPHONE_VOL_MAX;
			lhvol = tlv320aic_getreg(TLV_LHVOL);
			if (lval != (lhvol & TLV_HEADPHONE_VOL_MASK)) {
				lhvol =
				  (lhvol & ~TLV_HEADPHONE_VOL_MASK) | lval;
				tlv320aic_setreg(TLV_LHVOL, lhvol);
				change = 1;
			}

			rval = right + TLV_HEADPHONE_VOL_MIN;
			if (rval > TLV_HEADPHONE_VOL_MAX)
				rval = TLV_HEADPHONE_VOL_MAX;
			rhvol = tlv320aic_getreg(TLV_RHVOL);
			if (rval != (rhvol & TLV_HEADPHONE_VOL_MASK)) {
				rhvol =
				  (rhvol & ~TLV_HEADPHONE_VOL_MASK) | rval;
				tlv320aic_setreg(TLV_RHVOL, rhvol);
				change = 1;
			}
			break;
		}
	default:
		return -ENODEV;
	}
	
	return change;
}

static int
sia_audio_get_input_source(sia_t *sia)
{
	u16 apath_insel;

	DEBUG_FUNC();

	apath_insel = (tlv320aic_getreg(TLV_APATH) & 0x004);
	return (apath_insel ? 1 : 0);
}

static int
sia_audio_set_input_source(sia_t *sia, int source)
{
	u16 apath;

	DEBUG_FUNC();

	apath = (tlv320aic_getreg(TLV_APATH) & ~0x004);
	apath |= (source ? 0x004 : 0x000);
	tlv320aic_setreg(TLV_APATH, apath);
	return 0;
}

static int
sia_audio_set_sampling_rate(sia_t *sia, unsigned int sampling_rate)
{
	static unsigned int previous = 0;
	u16 rate;
	u16 reg;

	DEBUG_FUNC();

	if (sampling_rate == previous)
		return 0;

	switch (sampling_rate) {
	case 96000:
		rate = 0x01c;
		break;
	case 48000:
		rate = 0x000;
		break;
	case 32000:
		rate = 0x018;
		break;
	case 8000:
		rate = 0x00c;
		break;
	default:
		return -1;
	}

	reg = tlv320aic_getreg(TLV_SRATE);
	reg = (reg & ~0x03f) | rate;
	tlv320aic_setreg(TLV_SRATE, reg);

	previous = sampling_rate;

	codec_dump_register();

	return 0;
}

static int
sia_audio_set_format(sia_t *sia, snd_pcm_format_t format)
{
	static snd_pcm_format_t previous = 0;

	DEBUG_FUNC();

	if (format == previous)
		return 0;

	switch (format) {
	case SNDRV_PCM_FORMAT_S16_BE:
		sia->control &= ~SIA_CONTROL_BYTE_SWAP;
		break;
	case SNDRV_PCM_FORMAT_S16_LE:
		sia->control |= SIA_CONTROL_BYTE_SWAP;
		break;
	default:
		return -1;
	}

	out_be32(REG_SIA_CONTROL, sia->control);

	previous = format;

	return 0;
}

static int __devinit
sia_audio_init_i2s_controller(sia_t *sia)
{
	DEBUG_FUNC();

	sia->control = SIA_CONTROL_PLAY_RST | SIA_CONTROL_REC_RST;
	sia->interrupt_enable = 0;
	out_be32(REG_SIA_CONTROL, sia->control);
	out_be32(REG_SIA_INTERRUPT_ENABLE, sia->interrupt_enable);
	out_be32(REG_SIA_MONO_SELECT, SIA_STEREO);

	i2s_dump_register();

	return 0;
}

static int __devinit
sia_audio_init_codec(sia_t *sia)
{
	DEBUG_FUNC();

	tlv320aic_setreg(TLV_RESET, 0);

	/* default sampling rate: 48k */
	tlv320aic_setreg(TLV_LIVOL,  0x00);
	tlv320aic_setreg(TLV_RIVOL,  0x00);
	tlv320aic_setreg(TLV_LHVOL,  0x30);
	tlv320aic_setreg(TLV_RHVOL,  0x30);
	tlv320aic_setreg(TLV_APATH,  0x10);/* 14: INPUT:Mic, 10: INPUT:Line */
	tlv320aic_setreg(TLV_DPATH,  0x00);
	tlv320aic_setreg(TLV_PWR,    0x00);
	tlv320aic_setreg(TLV_DIFORM, 0x42);
	tlv320aic_setreg(TLV_SRATE,  0x00);
	tlv320aic_setreg(TLV_DACT,   0x01);

	return 0;
}

static int __devinit
sia_audio_init(sia_t *sia)
{
	DEBUG_FUNC();

	sia_audio_init_i2s_controller(sia);
	sia_audio_init_codec(sia);

	sia_audio_set_format(sia, DEFAULT_FORMAT);
	sia_audio_set_sampling_rate(sia, DEFAULT_SAMPLING_RATE);
	sia_audio_set_volume(sia, MIXERID_LINE_INPUT_VOL,
			     DEFAULT_LINE_INPUT_VOLUME);
	sia_audio_set_volume(sia, MIXERID_HEADPHONE_VOL,
			     DEFAULT_HEADPHONE_VOLUME);

	return 0;
}

static int
sia_audio_shutdown(sia_t *sia)
{
	DEBUG_FUNC();

	return 0;
}

static int
sia_mixer_volume_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo)
{
	sia_t *sia = kcontrol->private_data;
	int mixerid = kcontrol->private_value;

	DEBUG_FUNC();

	return sia_audio_get_volume_info(sia, mixerid, uinfo);
}

static int
sia_mixer_volume_get(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *uvalue)
{
	sia_t *sia = kcontrol->private_data;
	int mixerid = kcontrol->private_value;
	int val;

	DEBUG_FUNC();

	val = sia_audio_get_volume(sia, mixerid);
	if (val < 0)
		return val;

	uvalue->value.integer.value[0] = val & 0xff;
	uvalue->value.integer.value[1] = val >> 8;

	return 0;
}

static int
sia_mixer_volume_put(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *uvalue)
{
	sia_t *sia = kcontrol->private_data;
	int mixerid = kcontrol->private_value;
	int val;

	DEBUG_FUNC();

	val = (uvalue->value.integer.value[0] | 
	       (uvalue->value.integer.value[1] << 8));

	return sia_audio_set_volume(sia, mixerid, val);
}

static snd_kcontrol_new_t sia_line_input_vol_ctrl = {
	.iface		= SNDRV_CTL_ELEM_IFACE_MIXER,
	.name		= "Line Playback Volume",
	.index		= 0x00,
	.info		= sia_mixer_volume_info,
	.get		= sia_mixer_volume_get,
	.put		= sia_mixer_volume_put,
	.private_value	= MIXERID_LINE_INPUT_VOL,
};

static snd_kcontrol_new_t sia_headphone_vol_ctrl = {
	.iface		= SNDRV_CTL_ELEM_IFACE_MIXER,
	.name		= "Headphone Playback Volume",
	.index		= 0x00,
	.info		= sia_mixer_volume_info,
	.get		= sia_mixer_volume_get,
	.put		= sia_mixer_volume_put,
	.private_value	= MIXERID_HEADPHONE_VOL,
};

static int
sia_mixer_input_source_info(snd_kcontrol_t *kcontrol,
			    snd_ctl_elem_info_t *uinfo)
{
	static char *label[2] = {
		"Line", "Mic"
	};

	DEBUG_FUNC();

	uinfo->type = SNDRV_CTL_ELEM_TYPE_ENUMERATED;
	uinfo->count = 1;
	uinfo->value.enumerated.items = 2;
	if (uinfo->value.enumerated.item >= uinfo->value.enumerated.items) {
		uinfo->value.enumerated.item =
			uinfo->value.enumerated.items - 1;
	}
	strcpy(uinfo->value.enumerated.name,
	       label[uinfo->value.enumerated.item]);
	return 0;
}

static int
sia_mixer_input_source_get(snd_kcontrol_t *kcontrol,
			    snd_ctl_elem_value_t *uvalue)
{
	sia_t *sia = kcontrol->private_data;

	DEBUG_FUNC();

	uvalue->value.enumerated.item[0] = sia_audio_get_input_source(sia);
	return 0;
}

static int
sia_mixer_input_source_put(snd_kcontrol_t *kcontrol,
			    snd_ctl_elem_value_t *uvalue)
{
	sia_t *sia = kcontrol->private_data;

	DEBUG_FUNC();

	sia_audio_set_input_source(sia, uvalue->value.enumerated.item[0]);
	return 0;
}

static snd_kcontrol_new_t sia_input_source_ctrl = {
	.iface		= SNDRV_CTL_ELEM_IFACE_MIXER,
	.name		= "Input Source",
	.count		= 1,
	.info		= sia_mixer_input_source_info,
	.get		= sia_mixer_input_source_get,
	.put		= sia_mixer_input_source_put,
};

static int __devinit
sia_mixer_new(snd_card_t *card)
{
	snd_kcontrol_t *kctl;
	int ret;

	DEBUG_FUNC();

	kctl = snd_ctl_new1(&sia_line_input_vol_ctrl, NULL);
	kctl->private_data = card->private_data;
	ret = snd_ctl_add(card, kctl);
	if (ret < 0)
		return ret;

	kctl = snd_ctl_new1(&sia_headphone_vol_ctrl, NULL);
	kctl->private_data = card->private_data;
	ret = snd_ctl_add(card, kctl);
	if (ret < 0)
		return ret;

	kctl = snd_ctl_new1(&sia_input_source_ctrl, NULL);
	kctl->private_data = card->private_data;
	ret = snd_ctl_add(card, kctl);
	if (ret < 0)
		return ret;

	return 0;
}

static snd_pcm_hardware_t sia_hw_info = {
	.info             = (SNDRV_PCM_INFO_INTERLEAVED | 
			     SNDRV_PCM_INFO_BLOCK_TRANSFER),
	.formats          = (SNDRV_PCM_FMTBIT_S16_BE |
			     SNDRV_PCM_FMTBIT_S16_LE),
	.rates            = (SNDRV_PCM_RATE_8000 |
			     SNDRV_PCM_RATE_32000 |
			     SNDRV_PCM_RATE_48000 |
			     SNDRV_PCM_RATE_96000),
	.rate_min         = 8000,
	.rate_max         = 96000,
	.channels_min     = 2,
	.channels_max     = 2,
	.buffer_bytes_max = 64 * 1024,
	.period_bytes_min = 256,
	.period_bytes_max = PAGE_SIZE,
	.periods_min      = 4,
	.periods_max      = PAGE_SIZE / 16,
};

static int
sia_pcm_open(snd_pcm_substream_t *substream)
{
	sia_t *sia = substream->private_data;
	sia_runtime_t *siarun;
	snd_pcm_runtime_t *runtime = substream->runtime;

	DEBUG_FUNC();

	switch (substream->stream) {
	case SNDRV_PCM_STREAM_PLAYBACK:
		siarun = &sia->playback;
		break;
	case SNDRV_PCM_STREAM_CAPTURE:
		siarun = &sia->capture;
		break;
	default:
		return -EINVAL;
	}

	siarun->substream = substream;
	runtime->private_data = siarun;
	runtime->hw = sia_hw_info;

	sia->control &= ~siarun->control_fifo_reset_bit;
	out_be32(REG_SIA_CONTROL, sia->control);

	return 0;
}

static int
sia_pcm_close(snd_pcm_substream_t *substream)
{
	sia_t *sia = substream->private_data;
	sia_runtime_t *siarun = substream->runtime->private_data;

	DEBUG_FUNC();

	WARN_ON(sia->interrupt_enable & siarun->interrupt_enable_bit);

	sia->control |= siarun->control_fifo_reset_bit;
	out_be32(REG_SIA_CONTROL, sia->control);

	siarun->substream = NULL;

	return 0;
}

static void
sia_dma_hw_free(snd_pcm_substream_t *substream)
{
	sia_t *sia = substream->private_data;
	snd_pcm_runtime_t *runtime = substream->runtime;
	sia_runtime_t *siarun = runtime->private_data;
	struct snd_dma_buffer *buf = runtime->dma_buffer_p;

	DEBUG_FUNC();

	if (runtime->dma_area == NULL)
		return;

	siarun->start	= NULL;

	if (buf != &substream->dma_buffer) {
		dma_free_coherent(sia->dev, buf->bytes, buf->area, buf->addr);
		kfree(runtime->dma_buffer_p);
	}

	snd_pcm_set_runtime_buffer(substream, NULL);
}

static int
sia_dma_hw_alloc(snd_pcm_substream_t *substream, size_t size)
{
	sia_t *sia = substream->private_data;
	snd_pcm_runtime_t *runtime = substream->runtime;
	struct snd_dma_buffer *buf = runtime->dma_buffer_p;

	DEBUG_FUNC();

	if (buf) {
		if (buf->bytes >= size)
			goto out;
		sia_dma_hw_free(substream);
	}

	if (substream->dma_buffer.area != NULL &&
	    substream->dma_buffer.bytes >= size) {
		buf = &substream->dma_buffer;
	}
	else {
		buf = kmalloc(sizeof(struct snd_dma_buffer), GFP_KERNEL);
		if (!buf)
			goto nomem;

		buf->dev.type = SNDRV_DMA_TYPE_DEV;
		buf->dev.dev = sia->dev;
		buf->area = dma_alloc_coherent(sia->dev, size, &buf->addr,
					       GFP_KERNEL);
		if (!buf->area)
			goto free;
		buf->bytes = size;
		buf->private_data = NULL;
	}
	snd_pcm_set_runtime_buffer(substream, buf);
 out:
	runtime->dma_bytes = size;
	return 0;

 free:
	kfree(buf);
 nomem:
	return -ENOMEM;
}

static int
sia_pcm_hw_free(snd_pcm_substream_t *substream)
{
	sia_t *sia = substream->private_data;
	sia_runtime_t *siarun = substream->runtime->private_data;

	DEBUG_FUNC();

	WARN_ON(sia->interrupt_enable & siarun->interrupt_enable_bit);

	sia_dma_hw_free(substream);

	return 0;
}

static int
sia_pcm_hw_params(snd_pcm_substream_t *substream,
		  snd_pcm_hw_params_t *params)
{
	DEBUG_FUNC();

	sia_pcm_hw_free(substream);

	return sia_dma_hw_alloc(substream, params_buffer_bytes(params));
}

static int
sia_pcm_prepare(snd_pcm_substream_t *substream)
{
	sia_t *sia = substream->private_data;
	snd_pcm_runtime_t *runtime = substream->runtime;
	sia_runtime_t *siarun = runtime->private_data;

	DEBUG_FUNC();

	pr_debug("runtime:rate        = %d\n", runtime->rate);
	pr_debug("runtime:frame_bits  = %d\n", runtime->frame_bits);
	pr_debug("runtime:channel     = %d\n", runtime->channels);

	if (sia_audio_set_format(sia, runtime->format) < 0)
		return -EINVAL;
	if (sia_audio_set_sampling_rate(sia, runtime->rate) < 0)
		return -EINVAL;

	siarun->start	= (void *)runtime->dma_area;
	siarun->end	= siarun->start + runtime->dma_bytes;
	siarun->ptr	= siarun->start;
	siarun->period	= siarun->bytes	=
			  frames_to_bytes(runtime, runtime->period_size);

	return 0;
}

static void
sia_pcm_start(snd_pcm_substream_t *substream)
{
	sia_t *sia = substream->private_data;
	sia_runtime_t *siarun = substream->runtime->private_data;

	DEBUG_FUNC();

	sia->interrupt_enable |= siarun->interrupt_enable_bit;
	out_be32(REG_SIA_INTERRUPT_ENABLE, sia->interrupt_enable);
}

static void
sia_pcm_stop(snd_pcm_substream_t *substream)
{
	sia_t *sia = substream->private_data;
	sia_runtime_t *siarun = substream->runtime->private_data;

	DEBUG_FUNC();

	sia->interrupt_enable &= ~siarun->interrupt_enable_bit;
	out_be32(REG_SIA_INTERRUPT_ENABLE, sia->interrupt_enable);
}

static int
sia_pcm_trigger(snd_pcm_substream_t *substream, int cmd)
{
	sia_t *sia = substream->private_data;
	unsigned long flags;
	int ret = 0;

	DEBUG_FUNC();

	spin_lock_irqsave(&sia->lock, flags);
	switch (cmd) {
	case SNDRV_PCM_TRIGGER_START:
		sia_pcm_start(substream);
		break;

	case SNDRV_PCM_TRIGGER_STOP:
		sia_pcm_stop(substream);
		break;

	default:
		ret = -EINVAL;
	}
	spin_unlock_irqrestore(&sia->lock, flags);

	return ret;
}

static snd_pcm_uframes_t
sia_pcm_pointer(snd_pcm_substream_t *substream)
{
	snd_pcm_runtime_t *runtime = substream->runtime;
	sia_runtime_t *siarun = runtime->private_data;
	ssize_t bytes = siarun->ptr - siarun->start;

	DEBUG_FUNC();

	return bytes_to_frames(runtime, bytes);
}

static snd_pcm_ops_t sia_pcm_ops = {
	.open		= sia_pcm_open,
	.close		= sia_pcm_close,
	.ioctl		= snd_pcm_lib_ioctl,
	.hw_params	= sia_pcm_hw_params,
	.hw_free	= sia_pcm_hw_free,
	.prepare	= sia_pcm_prepare,
	.trigger	= sia_pcm_trigger,
	.pointer	= sia_pcm_pointer,
};

static int __devinit
sia_pcm_new(snd_card_t *card)
{
	snd_pcm_t *pcm;
	int play = 1;
	int capt = 1;
	int ret;

	DEBUG_FUNC();

	ret = snd_pcm_new(card, "SIA-PCM", 0, play, capt, &pcm);
	if (ret)
		return ret;

	((sia_t *)card->private_data)->pcm = pcm;
	pcm->private_data = card->private_data;
	pcm->private_free = NULL;

	strlcpy(pcm->name, DRIVER_NAME, sizeof(pcm->name));

	/* alsa pcm ops setting for SNDRV_PCM_STREAM_PLAYBACK */
	if (play)
		snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK, &sia_pcm_ops);

	/* alsa pcm ops setting for SNDRV_PCM_STREAM_CAPTURE */
	if (capt)
		snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE, &sia_pcm_ops);

	return 0;
}

static int __devinit 
sia_probe(struct platform_device *pdev)
{
	int irq = platform_get_irq(pdev, 0);
	struct resource *res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	u32 tmp;
	u16 board_id, board_version;
	snd_card_t *card;
	sia_t *sia;
	int ret;

	DEBUG_FUNC();

	if (irq < 0) {
		dev_err(&pdev->dev, "no irq for device\n");
		return -ENOENT;
	}
	if (!res) {
		dev_err(&pdev->dev, "no mem for device\n");
		return -ENOENT;
	}

	if (!request_mem_region(res->start, res->end - res->start + 1,
				DRIVER_NAME)) {
		return -EBUSY;
	}
	sia_addr = (void __iomem *)ioremap_nocache(res->start,
						   res->end - res->start + 1);
	if (!sia_addr) {
		ret = -EBUSY;
		goto release_mem;
	}

	tmp = in_be32(REG_SIA_CORE_ID);
	board_id = (tmp >> 16) & 0xffff;
	board_version = tmp & 0xffff;

	pr_info("  OPB-SIA00 (id: %#hx; ver: %#hx)\n",
		board_id, board_version);

	if (!is_known_board(board_id, board_version)) {
		dev_err(&pdev->dev, "  Unsupported ID and/or Version\n");
		ret = -ENODEV;
		goto unmap_io;
	}

	card = snd_card_new(SNDRV_DEFAULT_IDX1, SNDRV_DEFAULT_STR1,
			    THIS_MODULE, sizeof(sia_t));
	if (card == NULL) {
		ret = -ENOMEM;
		goto unmap_io;
	}

	strlcpy(card->driver,    DRIVER_NAME, sizeof(card->driver));
	strlcpy(card->shortname, DRIVER_NAME, sizeof(card->shortname));
	strlcpy(card->longname,  DRIVER_NAME, sizeof(card->longname));

	sia = card->private_data;
	spin_lock_init(&sia->lock);
	sia->dev = &pdev->dev;

	sia->playback.control_fifo_reset_bit = SIA_CONTROL_PLAY_RST;
	sia->playback.interrupt_enable_bit = SIA_TX_FIFO_HEMP;
	sia->playback.fifosz = in_be32(REG_SIA_AUDIO_TX_FIFO_DEPTH) * 4;
	if (!sia->playback.fifosz) {
		ret = -ENODEV;
		goto nodev;
	}
	sia_hw_info.fifo_size = sia->playback.fifosz;

	sia->capture.control_fifo_reset_bit = SIA_CONTROL_REC_RST;
	sia->capture.interrupt_enable_bit = SIA_RX_FIFO_HFULL;
	sia->capture.fifosz = in_be32(REG_SIA_AUDIO_RX_FIFO_DEPTH) * 4;
	if (!sia->capture.fifosz) {
		ret = -ENODEV;
		goto nodev;
	}
	if (sia_hw_info.fifo_size < sia->capture.fifosz)
		sia_hw_info.fifo_size = sia->capture.fifosz;

	ret = sia_audio_init(sia);
	if (ret < 0)
		goto unmap_io;

	ret = request_irq(irq, sia_irq, IRQF_DISABLED, DRIVER_NAME, sia);
	if (ret)
		goto nodev;

	ret = sia_pcm_new(card);
	if (ret < 0)
		goto release_irq;

	ret = sia_mixer_new(card);
	if (ret < 0)
		goto release_irq;

	snd_card_set_dev(card, &pdev->dev);

	ret = snd_card_register(card);
	if (ret < 0)
		goto release_irq;

	platform_set_drvdata(pdev, card);
	return 0;

 release_irq:
	free_irq(irq, sia);
 nodev:
	snd_card_free(card);
 unmap_io:
	iounmap(sia_addr);
 release_mem:
	release_mem_region(res->start, res->end - res->start + 1);
	return ret;
}

static int __devexit
sia_remove(struct platform_device *pdev)
{
	int irq = platform_get_irq(pdev, 0);
	struct resource *res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	snd_card_t *card = platform_get_drvdata(pdev);
	sia_t *sia;

	DEBUG_FUNC();

	if (card) {
		sia = card->private_data;
		sia_audio_shutdown(sia);

		free_irq(irq, sia);

		snd_card_free(card);
		platform_set_drvdata(pdev, NULL);
	}

	iounmap(sia_addr);
	release_mem_region(res->start, res->end - res->start + 1);

	return 0;
}

static struct platform_driver sia_driver = {
	.probe		= sia_probe,
	.remove		= __devexit_p(sia_remove),
	.driver		= {
		.name	= DRIVER_NAME,
	},
};

static int __init 
sia_init(void)
{
	DEBUG_FUNC();

	printk("%s (v%s) : %s\n", DRIVER_NAME, DRIVER_VERSION, DRIVER_DESC);

	return platform_driver_register(&sia_driver);
}

static void __exit 
sia_exit(void)
{
	DEBUG_FUNC();

	platform_driver_unregister(&sia_driver);
}

module_init(sia_init);
module_exit(sia_exit);

MODULE_VERSION(DRIVER_VERSION);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_LICENSE("GPL");
